#   Copyright (c) 2010, 2011 Patricio Paez <pp@pp.com.mx>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>

 # We are doing math in this module ...


__doc__ = '''
arithmetic is a Python module that allows mixing arithmetic
operations and text.

Usage as command:

      arithmetic expression(s) Evaluate expression(s).
      arithmetic -f path       Read path contents and evaluate.
      arithmetic               Read from standard input and evaluate.
      arithmetic -h            Show usage

Examples:

      $ arithmetic 2 x 3
      6
      $ echo 4 x 12 | arithmetic
      48
      $ echo 'a = 5    a + 4 =' | arithmetic
      a = 5    a + 4 = 9
      $ arithmetic < filename
      ...

Usage as module:

     from arithmetic import feed
     ...
     resultText = feed( inputText )
     ...
'''

import logging

logger = logging.getLogger('zim.inc.arithmetic')


import re


renumber = re.compile(r'(-?[0-9][0-9,]*(\.[0-9]*)?%?)|(\.[0-9]+%?)')
#reidentifier = re.compile( r'[a-zA-Z][a-zA-Z0-9_]*' )
#rexenclosed = re.compile( r'[0-9.)](x)[^a-zA-Z]' )
reidentifier = re.compile(r'[^\W\d]\w*', re.U) # letter (but not a number) followed by alpha-numeric
rexenclosed = re.compile(r'[0-9.)](x)[\W\d]', re.U)

class Lexer:
    ''
    def __init__(self, text):
        self.text = text
        self.offset = 0

    def gettoken(self):
        '''Get next token from text and return its type and value.

        Return (None,None) if no more text.

        Identifiers    letter ( letter | digit | _ )*
        Numbers        digit ( digit | , | . ) *
        Operators      - + * / ^ ** x

        x can be a name or an operator

        x preceded and followed by a digit is taken as if
        preceded and followed by a space.  i.e 5x3
        is seen as 5 x 3.
        '''
        while self.offset < len(self.text):
            if self.text[self.offset] == ' ':
                value = self.text[self.offset]
                self.offset = self.offset + 1
                continue
            if self.text[self.offset] == '\n':
                value = self.text[self.offset]
                self.offset = self.offset + 1
                self.type = 'r'
                self.value = value
                return
            m = renumber.match(self.text, self.offset)
            if m:
                value = m.group()
                self.offset = m.end()
                self.type = 'f'
                self.value = value
                return
            m = rexenclosed.match(self.text, self.offset - 1)
            if m:
                value = m.group(1)
                self.offset = m.end(1)
                self.type = 'x'
                self.value = value
                return
            m = reidentifier.match(self.text, self.offset)
            if m:
                value = m.group()
                self.offset = m.end()
                self.type = 'n'
                self.value = value
                return
            if self.text[self.offset: self.offset + 2] == '**':
                value = self.text[self.offset: self.offset + 2]
                self.offset = self.offset + 2
                self.type = 'o'
                self.value = value
                return
            if self.text[self.offset] in '+-*/^()':
                value = self.text[self.offset]
                self.offset = self.offset + 1
                self.type = 'o'
                self.value = value
                return
            value = self.text[self.offset] + '***'
            self.offset = self.offset + 1
            self.type = 'u'
            self.value = value
            return
        self.type = ''
        self.value = None


from decimal import Decimal, getcontext
getcontext().prec = 100


def safe_eval(expression):
    '''Safe evaluation of a python expression'''
    GLOBALS = {'__builtins__': None} # Don't allow open() etc.
    try:
        re = eval(expression, GLOBALS, {'Decimal': Decimal})
    except Exception as e:
        raise

    return re


def evaluate(expression_text, UseDigitGrouping = True, variables = {}, functions = {}):
    '''Parse expression, calculate and return its result.

    if UseDigitGrouping is True, the result includes commas.

    '''
    expression = []

    lexer = Lexer(expression_text)

    def factor():
        if lexer.value in '-+':
            expression.append(lexer.value)
            lexer.gettoken()
        if lexer.type == 'f':
            #  remove commas and spaces
            #  translate % to /100'''
            value = lexer.value.replace('%', '')
            value = value.replace(' ', '')
            expression.append(value.replace(',', ''))
            if lexer.value[-1] == '%':
                expression.append('/')
                expression.append('100')
            lexer.gettoken()
        elif lexer.value == '(':
            expression.append(lexer.value)
            lexer.gettoken()
            expr()
            expression.append(lexer.value)
            lexer.gettoken()
        elif lexer.type == 'n':
            name = lexer.value
            lexer.gettoken()
            # handle multiple word names
            while lexer.type == 'n' and lexer.value != 'x':
                name = name + ' ' + lexer.value
                lexer.gettoken()
            if name in variables:
                # remove commas
                value = variables[name].replace(',', '')
                expression.append(value)
            elif name in functions:
                if name not in functions[name]:
                    # standard formula
                    expression.append(str(evaluate(functions[name],
                                       UseDigitGrouping = False,
                                       variables=variables, functions=functions)))
                else:
                    # recurrent relation without initial value
                    expression.append('0')
            else:
                expression.append(name + ' undefined')

    def factors():
        if lexer.value == '^' or lexer.value == '**':
            expression.append('**')
            lexer.gettoken()
            power()

    def power():
        factor()
        factors()

    def powers():
        if lexer.value and lexer.value in '*x/':
            expression.append(lexer.value.replace('x', '*'))
            lexer.gettoken()
            power()
            powers()

    def term():
        power()
        powers()

    def terms():
        if lexer.value and lexer.value in '-+':
            expression.append(lexer.value)
            lexer.gettoken()
            term()
            terms()

    def expr():
        term()
        terms()

    lexer.gettoken()
    expr()

    expressionD = []
    for element in expression:
        if element not in '+-x/^*()' and element != '**':
            element = "Decimal('" + element + "')"
        expressionD.append(element)

    if UseDigitGrouping:
        return AddCommas(safe_eval(''.join(expressionD)))
    else:
        return safe_eval(''.join(expressionD))



reEqualSign = re.compile(' ?= ?')
reSepar = re.compile('  +|\t')
reColonLeft = re.compile(': ')



def TypeAndValueOf(expression):
    '''"Returns a (type, value) tuple.

    type may be:

    v = void
    i = integer, f = float
    e = expression containing variables
    a = arithmetic expression, no variables
    n = name

    value is expression with some modifications:
    blank spaces and commas removed, x replaced by *.'''

    lexer = Lexer(expression)

    # Capture the odd token types
    lexer.text = re.sub('[()]', '', lexer.text)
    lexer.gettoken()
    oddtokens = lexer.type
    while lexer.type:
        lexer.gettoken()
        lexer.gettoken()
        oddtokens = oddtokens + lexer.type

    if oddtokens == '':  # empty
        return 'v', ''
    elif oddtokens == 'f':
        return 'f', expression
    elif oddtokens == 'n':
        return 'n', expression
    elif 'n' in oddtokens or 'x' in oddtokens:
        return 'e', expression  # expression with names
    else:
        return 'a', expression

class Parser:
    'Base class'

    def __init__(self, output_decimals=6):
        # Written by parseLine(), read by evaluate():
        self.output_decimals = output_decimals
        self.functions = {}
        self.variables = {}

    def parse(self, text):
        'Find expressions in text, return it with results.'

        lines = text.splitlines()

        for i in range(self.countLines(lines)):
            self.parseLine(i, lines, variables=self.variables, functions=self.functions)

        return '\n'.join(lines)

    def countLines(self, lines):
        'Return number of lines.'
        return len(lines)

    def readLine(self, i, lines):
        'Return i line from lines.'
        return lines[i]

    def writeResult(self, i, lines, start, end, text):
        'Write text in line i of lines from start to end offset.'

        lines[i] = lines[i][:start] + text + lines[i][end:]

    def parseLine(self, i, lines, variables={}, functions={}):
            'Find and evaluate expressions in line i.'

            # get line
            line = self.readLine(i, lines)

            RightPrevStart = 0
            RightPrevEnd = 0
            mEqualSignPrev = re.search('^', line)
            mEqualSignAct = reEqualSign.search(line, mEqualSignPrev.end())
            while mEqualSignAct:

                # Determine LeftActStart,
                # the larger of mEqualSignPrev, mSeparLeft, mColonLeft, mBeginOfLine
                LeftStarts = []
                LeftStarts.append(mEqualSignPrev.end())

                mSeparLeft = reSepar.search(line, mEqualSignPrev.end(), mEqualSignAct.start())
                if mSeparLeft:
                    SeparLeftEnd = mSeparLeft.end()
                    mSeparLeft = reSepar.search(line, mSeparLeft.end(), mEqualSignAct.start())
                    while mSeparLeft:     # search next
                        SeparLeftEnd = mSeparLeft.end()
                        mSeparLeft = reSepar.search(line, mSeparLeft.end(), mEqualSignAct.start())
                    LeftStarts.append(SeparLeftEnd)

                mColonLeft = reColonLeft.search(line, mEqualSignPrev.end(), mEqualSignAct.start())
                if mColonLeft:
                    LeftStarts.append(mColonLeft.end())
                mBeginOfLine = re.search('^ *', line)
                LeftStarts.append(mBeginOfLine.end())
                LeftActStart = max(LeftStarts)

                # Determine RightActEnd,
                # the smaller of mEqualSignNext, mSeparRight, endofline
                RightEnds = []
                mEqualSignNext = reEqualSign.search(line, mEqualSignAct.end())
                if mEqualSignNext:
                    RightEnds.append(mEqualSignNext.start())
                mSeparRight = reSepar.search(line, mEqualSignAct.end())
                if mSeparRight:
                    RightEnds.append(mSeparRight.start())
                mEndOfLine = re.search(' *$', line)
                RightEnds.append(mEndOfLine.start())
                RightActEnd = min(RightEnds)

                rangolibre = line[RightPrevEnd: LeftActStart]
                rangoLeft = line[LeftActStart: mEqualSignAct.start()]
                rangocentro = line[mEqualSignAct.start(): mEqualSignAct.end()]
                rangoRight = line[mEqualSignAct.end(): RightActEnd]

                tipoLeft, valorLeft = TypeAndValueOf(rangoLeft)
                tipoRight, valorRight = TypeAndValueOf(rangoRight)

                if tipoLeft != 'v': # there is something to the left

                    # perform operations
                    if tipoLeft in 'eaif' and tipoRight in 'vif': # evaluate expression
                        try:
                            resultado = str(evaluate(valorLeft,
                                        variables=variables, functions=functions))
                            self.writeResult(i, lines, mEqualSignAct.end(), RightActEnd, resultado)
                        except:
                            logger.warning('eval error: %s, %s, %s, %s', tipoLeft, valorLeft, tipoRight, valorRight)
                    elif tipoLeft == 'n' and tipoRight in 'ifav':
                        if valorLeft not in functions:     # variable on the left
                            if tipoRight != 'v':    # assign to variable
                                try:
                                    variables[valorLeft] = str(evaluate(str(valorRight),
                                                            variables=variables, functions=functions))

                                except:
                                    logger.warning('exec error: %s, %s, %s, %s', tipoLeft, valorLeft, tipoRight, valorRight)
                                    raise
                            else:                   # evaluate a variable
                                if valorLeft in variables:
                                        resultado = variables[valorLeft]
                                        resultado = AddCommas(resultado)
                                        self.writeResult(i, lines, mEqualSignAct.end(), RightActEnd, resultado)

                        else:                                  # function on the left: evaluate
                            if valorLeft not in functions[valorLeft]:
                                try:                # standard formula
                                    resultado = str(evaluate(valorLeft,
                                                variables=variables, functions=functions))
                                    self.writeResult(i, lines, mEqualSignAct.end(), RightActEnd, resultado)
                                except:
                                    logger.warning('eval error: %s, %s, %s, %s', tipoLeft, valorLeft, tipoRight, valorRight)
                            else:                   # recurrence relation
                                if valorLeft not in variables:            # initial value
                                  if valorRight != '':
                                    variables[valorLeft] = str(evaluate(str(valorRight),
                                                            variables=variables, functions=functions))
                                else:                                         # iteration
                                    resultado = str(evaluate(functions[valorLeft],
                                                    variables=variables, functions=functions))
                                    self.writeResult(i, lines, mEqualSignAct.end(), RightActEnd, resultado)
                                    variables[valorLeft] = resultado

                    elif tipoLeft == 'n' and tipoRight in 'e': # define a function
                            functions[valorLeft] = str(valorRight)

                    elif tipoLeft == 'n' and tipoRight in 'n': # define an alias
                            functions[valorLeft] = str(valorRight)



                    RightPrevStart = mEqualSignAct.end()
                    RightPrevEnd = RightActEnd

                # get line
                line = self.readLine(i, lines)

                if mEqualSignNext:
                    mEqualSignNext = reEqualSign.search(line, mEqualSignAct.end())
                mEqualSignPrev = mEqualSignAct
                mEqualSignAct = mEqualSignNext


class ParserTk(Parser):
    ''

    def parse(self, textWidget):
        ''
        for i in range(self.countLines(textWidget)):
            self.parseLine(i, textWidget, variables=self.variables, functions=self.functions)

    def countLines(self, textWidget):
        ''
        return int(textWidget.index('end').split('.')[0])

    def readLine(self, i, textWidget):
        ''
        return textWidget.get(str(i) + '.0', str(i) + '.end')

    def writeResult(self, i, textWidget, start, end, text):
        'Write text in line i of lines from start to end offset.'
        textWidget.delete(str(i) + '.' + str(start), str(i) + '.' + str(end))
        textWidget.insert(str(i) + '.' + str(start), text)

class ParserGTK(Parser):
    ''

    def parse(self, textBuffer):
        ''
        for i in range(self.countLines(textBuffer)):
            self.parseLine(i, textBuffer, variables=self.variables, functions=self.functions)

    def countLines(self, textBuffer):
        ''
        return textBuffer.get_line_count()

    def readLine(self, i, textBuffer):
        ''
        iter_start = textBuffer.get_iter_at_line(i)
        if iter_start.ends_line():
            return ''
        else:
            iter_end = textBuffer.get_iter_at_line(i)
            iter_end.forward_to_line_end()
            return iter_start.get_text(iter_end)

    def writeResult(self, i, textBuffer, start, end, text):
        'Write text in line i of lines from start to end offset.'
        # Format decimals
        formatted = f"{float(text):.{self.output_decimals}f}"
        # Don't add unnecessary zeros
        if len(formatted) < len(text):
            text = formatted

        # Delete
        if end > start:
            # handle start at end of line or beyond
            iter_line = textBuffer.get_iter_at_line(i)
            nchars = iter_line.get_chars_in_line()
            if start > nchars - 1:
                start = nchars - 1
            iter_start = textBuffer.get_iter_at_line_offset(i, start)
            iter_end = textBuffer.get_iter_at_line_offset(i, end)
            textBuffer.delete(iter_start, iter_end)

        # Insert
        iter_start = textBuffer.get_iter_at_line_offset(i, start)
        textBuffer.insert(iter_start, text)

class ParserWx(Parser):
    ''

    def parse(self, TextControl):
        ''
        for i in range(self.countLines(TextControl)):
            self.parseLine(i, TextControl, variables=self.variables, functions=self.functions)

    def countLines(self, TextControl):
        ''
        return TextControl.GetNumberOfLines()

    def readLine(self, i, TextControl):
        ''
        return TextControl.GetLineText(i)

    def writeResult(self, i, TextControl, start, end, text):
        'Write text in line i of lines from start to end offset.'

        # Convert line, column to offset
        startOffset = TextControl.XYToPosition(start, i)
        endOffset = TextControl.XYToPosition(end, i)
        TextControl.Replace(startOffset, endOffset, text)

def feed(text):
    'Feed text to the parser.  It is processed line by line.'

    # Create instance of parser
    parser = Parser()

    # Parse the text
    return parser.parse(text)


def AddCommas(s):
    ''''Return s with thousands separators.

    Handles sign, decimals and thousands
    separator.
    '''

    s = str(s)
    s = s.replace(',', '')         #remove commas
    if s[0] in '-+':                #remove sign
            sign = s[0]
            s = s[1:]
    else:
            sign = ''
    if s[-1] == 'L':
        s = s[:-1]     #remove L suffix
    pos = s.find('.')
    if pos < 0:
        pos = len(s)
    while pos > 3:
            pos = pos - 3
            s = s[:pos] + ',' + s[pos:]
    s = sign + s                    #restore sign
    return s

if __name__ == '__main__':
    import sys
    if len(sys.argv) >= 2:
        if sys.argv[1] in ['-h', '--help']:
            print(__doc__)
        elif sys.argv[1] == '-f':
            filename = sys.argv[2]
            text = open(filename).read()
            print(feed(text))
        else:
            text = ' '.join(sys.argv[1:])
            if '=' in text:
                print(feed(text))
            else:
                print(evaluate(text))
    else:
        text = sys.stdin.read()
        lines = text.splitlines()
        if '=' in text or len(lines) > 1:
            print(feed(text))
        else:
            print(evaluate(lines[0]))




# Copyright 2008-2014 Jaap Karssenberg <jaap.karssenberg@gmail.com>

'''Module with basic filesystem objects.

This module must be used by all other zim modules for filesystem
interaction. It takes care of proper encoding file paths
(system dependent) and file contents (UTF-8) and implements a number
of sanity checks.

The main classes are L{File} and L{Dir} which implement file and
folder objects. There is also a singleton object to represent the whole
filesystem, whichprovides signals when a file or folder is created,
moved or deleted. This is stored in L{zim.fs.FS}.
'''

import os
import re
import sys
import shutil
import errno
import logging



from zim.errors import Error
from zim.parse.encode import url_decode, url_encode
from zim.signals import SignalEmitter, SIGNAL_AFTER

logger = logging.getLogger('zim.fs')


from zim.newfs.base import _os_expanduser, SEP, FileNotFoundError
from zim.newfs.local import AtomicWriteContext
from zim.newfs.local import get_tmpdir as _newfs_get_tmpdir

def adapt_from_newfs(file):
	from zim.newfs import LocalFile, LocalFolder

	if isinstance(file, LocalFile):
		return File(file.path)
	elif isinstance(file, LocalFolder):
		return Dir(file.path)
	else:
		return file


def adapt_from_oldfs(file):
	from zim.newfs import LocalFile, LocalFolder

	if isinstance(file, File):
		return LocalFile(file.path)
	elif isinstance(file, Dir):
		return LocalFolder(file.path)
	else:
		return file


try:
	from gi.repository import Gio
except ImportError:
	Gio = None

if not Gio:
	logger.info('No file monitor support - changes will go undetected')


xdgmime = None
mimetypes = None
if os.name == 'nt':
	# On windows even if xdg is installed, the database is not (always)
	# well initialized, so always fallback to mimetypes
	import mimetypes
else:
	try:
		import xdg.Mime as xdgmime
	except ImportError:
		logger.info("Can not import 'xdg.Mime' - falling back to 'mimetypes'")
		import mimetypes


#: Extensions to determine image mimetypes - used in L{File.isimage()}
IMAGE_EXTENSIONS = (
	# Gleaned from Gdk.get_formats()
	'bmp', # image/bmp
	'gif', # image/gif
	'icns', # image/x-icns
	'ico', # image/x-icon
	'cur', # image/x-icon
	'jp2', # image/jp2
	'jpc', # image/jp2
	'jpx', # image/jp2
	'j2k', # image/jp2
	'jpf', # image/jp2
	'jpeg', # image/jpeg
	'jpe', # image/jpeg
	'jpg', # image/jpeg
	'pcx', # image/x-pcx
	'png', # image/png
	'pnm', # image/x-portable-anymap
	'pbm', # image/x-portable-anymap
	'pgm', # image/x-portable-anymap
	'ppm', # image/x-portable-anymap
	'ras', # image/x-cmu-raster
	'tga', # image/x-tga
	'targa', # image/x-tga
	'tiff', # image/tiff
	'tif', # image/tiff
	'wbmp', # image/vnd.wap.wbmp
	'xbm', # image/x-xbitmap
	'xpm', # image/x-xpixmap
	'wmf', # image/x-wmf
	'apm', # image/x-wmf
	'svg', # image/svg+xml
	'svgz', # image/svg+xml
	'svg.gz', # image/svg+xml
	# Custom additions
	'webp', # image/webp
)


def isabs(path):
	'''Wrapper for C{os.path.isabs}.
	@param path: a file system path as string
	@returns: C{True} when the path is absolute instead of a relative path
	'''
	return path.startswith('file:/') \
	or path.startswith('~') \
	or os.path.isabs(path)


_tmpdir = None
def get_tmpdir():
	'''Get a folder in the system temp dir for usage by zim.
	This zim specific temp folder has permission set to be readable
	only by the current users, and is touched if it didn't exist yet.
	Used as base folder by L{TmpFile}.
	@returns: a L{Dir} object for the zim specific tmp folder
	'''
	global _tmpdir

	if _tmpdir is None:
		localdir = _newfs_get_tmpdir()
		_tmpdir = Dir(localdir.path)

	return _tmpdir


def normalize_file_uris(path):
	'''Function to deal with invalid or non-local file URIs.
	Translates C{file:/} to the proper C{file:///} form and replaces
	URIs of the form C{file://host/share} to C{smb://host/share}.
	@param path: a filesystem path or URL
	@returns: the proper URI or the original input path
	'''
	if path.startswith('file:///') \
	or path.startswith('file://localhost/'):
		return path
	elif path.startswith('file://'):
		return 'smb://' + path[7:]
	elif path.startswith('file:/'):
		return 'file:///' + path[6:]
	else:
		return path


def normalize_win32_share(path):
	r'''Translates paths for windows shares in the platform specific
	form. So on windows it translates C{smb://} URLs to C{\\host\share}
	form, and vice versa on all other platforms.
	Just returns the original path if it was already in the right form,
	or when it is not a path for a share drive.
	@param path: a filesystem path or URL
	@returns: the platform specific path or the original input path
	'''
	if os.name == 'nt':
		if path.startswith('smb://'):
			# smb://host/share/.. -> \\host\share\..
			path = path[4:].replace('/', '\\')
			path = url_decode(path)
	else:
		if path.startswith('\\\\'):
			# \\host\share\.. -> smb://host/share/..
			path = 'smb:' + url_encode(path.replace('\\', '/'))

	return path


def lrmdir(path):
	'''Wrapper for C{os.rmdir} that also knows how to unlink symlinks.
	Fails when the folder is not a link and is not empty.
	@param path: a file system path as string
	'''
	try:
		os.rmdir(path)
	except OSError:
		if os.path.islink(path) and os.path.isdir(path):
			os.unlink(path)
		else:
			raise


def cleanup_filename(name):
	'''Removes all characters in 'name' that are not allowed as part
	of a file name. This function is intended for e.g. config files etc.
	B{not} for page files in a store.
	For file system filenames we can not use:
	'\\', '/', ':', '*', '?', '"', '<', '>', '|'
	And we also exclude "\\t" and "\\n".
	@param name: the filename as string
	@returns: the name with invalid characters removed
	'''
	for char in ("/", "\\", ":", "*", "?", '"', "<", ">", "|", "\t", "\n"):
		name = name.replace(char, '')
	return name


def format_file_size(bytes):
	'''Returns a human readable label  for a file size
	E.g. C{1230} becomes C{"1.23kb"}, idem for "Mb" and "Gb"
	@param bytes: file size in bytes as integer
	@returns: size as string
	'''
	for unit, label in (
		(1000000000, 'Gb'),
		(1000000, 'Mb'),
		(1000, 'kb'),
	):
		if bytes >= unit:
			size = float(bytes) / unit
			if size < 10:
				return "%.2f%s" % (size, label)
			elif size < 100:
				return "%.1f%s" % (size, label)
			else:
				return "%.0f%s" % (size, label)
	else:
		return str(bytes) + 'b'




def _md5(content):
	import hashlib
	m = hashlib.md5()
	if isinstance(content, str):
		m.update(content.encode('UTF-8'))
	else:
		for l in content:
			m.update(l.encode('UTF-8'))
	return m.digest()


class PathLookupError(Error):
	'''Error raised when there is an error finding the specified path'''
	pass # TODO description


class FileWriteError(Error):
	'''Error raised when we can not write a file. Either due to file
	permissions or e.g. because it is detected the file changed on
	disk.
	'''
	pass # TODO description


class FileUnicodeError(Error):
	'''Error raised when there is an issue decoding the file contents.
	Typically due to different encoding where UTF-8 is expected.
	'''

	def __init__(self, file, error):
		self.file = file
		self.error = error
		self.msg = _('Could not read: %s') % file.path
			# T: message for FileUnicodeError (%s is the file name)
		self.description = _('This usually means the file contains invalid characters')
			# T: message for FileUnicodeError
		self.description += '\n\n' + _('Details') + ':\n' + str(error)
			# T: label for detailed error


# TODO actually hook the signal for deleting files and folders

class FSSingletonClass(SignalEmitter):
	'''Class used for the singleton 'zim.fs.FS' instance

	@signal: C{path-created (L{FilePath})}: Emitted when a new file or
	folder has been created
	@signal: C{path-moved (L{FilePath}, L{FilePath})}: Emitted when
	a file or folder has been moved
	@signal: C{path-deleted (L{FilePath})}: Emitted when a file or
	folder has been deleted

	@todo: fix the FS signals for folders as well
	'''

	# define signals we want to use - (closure type, return type and arg types)
	__signals__ = {
		'path-created': (SIGNAL_AFTER, None, (object,)),
		'path-moved': (SIGNAL_AFTER, None, (object, object)),
		'path-deleted': (SIGNAL_AFTER, None, (object,)),
	}


#: Singleton object for the system filesystem - see L{FSSingletonClass}
FS = FSSingletonClass()


class UnixPath(object):
	'''Base class for Dir and File objects, represents a file path

	@ivar path: the absolute file path as string
	file system encoding (should only be used by low-level functions)
	@ivar user_path: the absolute file path relative to the user's
	C{HOME} folder or C{None}
	@ivar uri: the C{file://} URI for this path
	@ivar basename: the basename of the path
	@ivar dirname: the dirname of the path
	@ivar dir: L{Dir} object for the parent folder

	@signal: C{changed (file, other_file, event_type)}: emitted when file
	changed - availability based on C{gio} support for file monitors on
	this platform
	'''

	def __init__(self, path):
		'''Constructor

		@param path: an absolute file path, file URL, L{FilePath} object
		or a list of path elements. When a list is given, the first
		element is allowed to be an absolute path, URL or L{FilePath}
		object as well.
		'''
		logger.warning('Using deprecated class "zim.fs.%s" - please update your code to use the "zim.newfs" module instead' % self.__class__.__name__)

		self._serialized = None

		if isinstance(path, FilePath):
			self.path = path.path
			return

		try:
			if isinstance(path, (list, tuple)):
				path = list(map(str, path))
					# Flatten objects - strings should be unicode or ascii already
				path = SEP.join(path)
					# os.path.join is too intelligent for it's own good
					# just join with the path separator.
			else:
				path = str(path) # make sure we can decode
		except UnicodeDecodeError:
			raise Error('BUG: invalid input, file names should be in ascii, or given as unicode')

		if path.startswith('file:/'):
			path = self._parse_uri(path)
		elif path.startswith('~'):
			path = _os_expanduser(path)

		self._set_path(path) # overloaded in WindowsPath

	def serialize_zim_config(self):
		'''Returns the file path as string for serializing the object'''
		if self._serialized is None:
			self._serialized = self.user_path or self.path
		return self._serialized

	@classmethod
	def new_from_zim_config(klass, string):
		'''Returns a new object based on the string representation for
		that path
		'''
		return klass(string)

	@staticmethod
	def _parse_uri(uri):
		# Spec is file:/// or file://host/
		# But file:/ is sometimes used by non-compliant apps
		# Windows uses file:///C:/ which is compliant
		if uri.startswith('file:///'):
			uri = uri[7:]
		elif uri.startswith('file://localhost/'):
			uri = uri[16:]
		elif uri.startswith('file://'):
			assert False, 'Can not handle non-local file uris'
		elif uri.startswith('file:/'):
			uri = uri[5:]
		else:
			assert False, 'Not a file uri: %s' % uri
		return url_decode(uri)

	def _set_path(self, path):
		self.path = os.path.abspath(path)

	def __iter__(self):
		parts = self.split()
		for i in range(1, len(parts)):
			path = os.path.join(*parts[0:i])
			yield Dir(path)

		#~ if self.isdir():
		yield Dir(self.path)
		#~ else:
			#~ yield self

	def __str__(self):
		return self.path

	def __repr__(self):
		return '<%s: %s>' % (self.__class__.__name__, self.path)

	def __add__(self, other):
		'''Concatenates paths, only creates objects of the same class. See
		L{Dir.file()} and L{Dir.subdir()} instead to create other objects.
		'''
		return self.__class__((self, other))

	def __eq__(self, other):
		return self.path == other.path

	def __ne__(self, other):
		return not self.__eq__(other)

	@property
	def basename(self):
		'''Basename property'''
		return os.path.basename(self.path) # encoding safe

	@property
	def dirname(self):
		'''Dirname property'''
		return os.path.dirname(self.path) # encoding safe

	@property
	def user_path(self):
		'''User_path property'''
		dir = Dir('~') # FIXME: Should we cache this folder somewhere ?
		if self.ischild(dir):
			return '~/' + self.relpath(dir)
		else:
			return None

	@property
	def uri(self):
		'''File uri property'''
		return 'file://' + url_encode(self.path)

	@property
	def dir(self):
		'''Returns a L{Dir} object for the parent dir'''
		path = os.path.dirname(self.path) # encoding safe
		return Dir(path)

	def monitor(self):
		'''Creates a L{FSObjectMonitor} for this path'''
		return FSObjectMonitor(self)

	def exists(self):
		'''Check if a file or folder exists.
		@returns: C{True} if the file or folder exists
		@implementation: must be implemented by sub classes in order
		that they enforce the type of the resource as well
		'''
		return os.path.exists(self.path)

	def iswritable(self):
		'''Check if a file or folder is writable. Uses permissions of
		parent folder if the file or folder does not (yet) exist.
		@returns: C{True} if the file or folder is writable
		'''
		if self.exists():
			return os.access(self.path, os.W_OK)
		else:
			return self.dir.iswritable() # recurs

	def mtime(self):
		'''Get the modification time of the file path.
		@returns: the mtime timestamp
		'''
		return os.stat(self.path).st_mtime

	def ctime(self):
		'''Get the creation time of the file path.
		@returns: the mtime timestamp
		'''
		return os.stat(self.path).st_ctime

	def size(self):
		'''Get file size in bytes
		See L{format_file_size()} to get a human readable label
		@returns: file size in bytes
		'''
		return os.stat(self.path).st_size

	def isequal(self, other):
		'''Check file paths are equal based on stat results (inode
		number etc.). Intended to detect when two files or dirs are the
		same on case-insensitive filesystems. Does not explicitly check
		the content is the same.
		If you just want to know if two files have the same content,
		see L{File.compare()}
		@param other: an other L{FilePath} object
		@returns: C{True} when the two paths are one and the same file
		'''
		# Do NOT assume paths are the same - could be hard link
		# or it could be a case-insensitive filesystem
		try:
			stat_result = os.stat(self.path)
			other_stat_result = os.stat(other.path)
		except OSError:
			return False
		else:
			return stat_result == other_stat_result

	def split(self):
		'''Split the parts of the path on the path separator.
		If the OS uses the concept of a drive the first part will
		include the drive. (So using split() to count the number of
		path elements will not be robust for the path "/".)
		@returns: a list of path elements
		'''
		drive, path = os.path.splitdrive(self.path)
		parts = path.replace('\\', '/').strip('/').split('/')
		parts[0] = drive + SEP + parts[0]
		return parts

	def relpath(self, reference, allowupward=False):
		'''Get a relative path for this file path with respect to
		another path. This method always returns paths using "/" as
		separator, even on windows.
		@param reference: a reference L{FilePath}
		@param allowupward: if C{True} the relative path is allowed to
		start with 'C{../}', if C{False} the reference should be a
		parent folder of this path.
		@returns: a relative file path
		@raises AssertionError: when C{allowupward} is C{False} and
		C{reference} is not a parent folder
		'''
		sep = SEP # '/' or '\'
		refdir = reference.path + sep
		if allowupward and not self.path.startswith(refdir):
			parent = self.commonparent(reference)
			if parent is None:
				return None # maybe on different drive under win32

			i = len(parent.path)
			j = refdir[i:].strip(sep).count(sep) + 1
			reference = parent
			path = '../' * j
		else:
			if not self.path.startswith(refdir):
				raise AssertionError('Not a parent folder')
			path = ''

		i = len(reference.path)
		path += self.path[i:].lstrip(sep).replace(sep, '/')
		return path

	def commonparent(self, other):
		'''Find a comon parent folder between two file paths.
		@param other: another L{FilePath}
		@returns: a L{Dir} object for the common parent folder, or
		C{None} when there is no common parent
		'''
		path = os.path.commonprefix((self.path, other.path)) # encoding safe
		path = path.replace(os.path.sep, SEP) # msys can have '/' as seperator
		i = path.rfind(SEP) # win32 save...
		if i >= 0:
			return Dir(path[:i + 1])
		else:
			# different drive ?
			return None

	def ischild(self, parent):
		'''Check if this path is a child path of a folder
		@returns: C{True} if this path is a child path of C{parent}
		'''
		return self.path.startswith(parent.path + SEP)

	def isdir(self):
		'''Check if this path is a folder or not. Used to detect if
		e.g. a L{File} object should have really been a L{Dir} object.
		@returns: C{True} when this path is a folder
		'''
		return os.path.isdir(self.path)

	def rename(self, newpath):
		'''Rename (move) the content this file or folder to another
		location. This will B{not} change the current file path, so the
		object keeps pointing to the old location.
		@param newpath: the destination C{FilePath} which can either be a
		file or a folder.
		@emits: path-moved
		'''
		# Using shutil.move instead of os.rename because move can cross
		# file system boundaries, while rename can not
		logger.info('Rename %s to %s', self, newpath)
		newpath = adapt_from_newfs(newpath)
		if self.path == newpath.path:
			raise AssertionError('Renaming %s to itself !?' % self.path)

		if newpath.isdir():
			if self.isequal(newpath):
				# We checked name above, so must be case insensitive file system
				# but we still want to be able to rename to other case, so need to
				# do some moving around
				tmpdir = self.dir.new_subdir(self.basename)
				shutil.move(self.path, tmpdir.path)
				shutil.move(tmpdir.path, newpath.path)
			else:
				# Needed because shutil.move() has different behavior for this case
				raise AssertionError('Folder already exists: %s' % newpath.path)
		else:
			# normal case
			newpath.dir.touch()
			shutil.move(self.path, newpath.path)

		FS.emit('path-moved', self, newpath)
		self.dir.cleanup()


class WindowsPath(UnixPath):
	'''Base class for Dir and File objects, represents a file path
	on windows.
	'''

	def _set_path(self, path):
		# Strip leading / for absolute paths
		if re.match(r'^[/\\]+[A-Za-z]:[/\\]', path):
			path = path.lstrip('/').lstrip('\\')
		self.path = os.path.abspath(path).replace('/', SEP) # msys can use '/' instead of '\\'

	@property
	def uri(self):
		'''File uri property with win32 logic'''
		# win32 paths do not start with '/', so add another one
		# and avoid url encoding the second ":" in "file:///C:/..."
		path = self.path.replace('\\', '/')
		if re.match('[A-Za-z]:/', path):
			return 'file:///' + path[:2] + url_encode(path[2:])
		else:
			return 'file:///' + url_encode(path)


# Determine which base class to use for classes below
if os.name == 'posix':
	FilePath = UnixPath
elif os.name == 'nt':
	FilePath = WindowsPath
else:
	logger.critical('os name "%s" unknown, falling back to posix', os.name)
	FilePath = UnixPath


class Dir(FilePath):
	'''Class representing a single file system folder'''

	def __eq__(self, other):
		if isinstance(other, Dir):
			return self.path == other.path
		else:
			return False

	def exists(self):
		return os.path.isdir(self.path)

	def list(self, glob=None, includehidden=False, includetmp=False, raw=False):
		'''List the file contents

		@param glob: a file name glob to filter the listed files, e.g C{"*.png"}
		@param includehidden: if C{True} include hidden files
		(e.g. names starting with "."), ignore otherwise
		@param includetmp: if C{True} include temporary files
		(e.g. names ending in "~"), ignore otherwise
		@param raw: for filtered folders (C{FilteredDir} instances)
		setting C{raw} to C{True} will disable filtering

		@returns: a sorted list of names for files and subdirectories.
		Will not return names that could not be decoded properly and
		will throw warnings if those are encountered.
		Hidden files are silently ignored.
		'''
		files = self._list(includehidden, includetmp)

		if glob:
			expr = _glob_to_regex(glob)
			files = list(filter(expr.match, files))

		files.sort()
		return files

	def _list(self, includehidden, includetmp):
		if self.exists():
			files = []
			for file in os.listdir(self.path):
				if file.startswith('.') and not includehidden:
					continue # skip hidden files
				elif (file.endswith('~') or file.startswith('~')) and not includetmp:
					continue # skip temporary files
				else:
					files.append(file)
			return files
		else:
			return []

	def walk(self, raw=True):
		'''Generator that yields all files and folders below this dir
		as objects.
		@param raw: see L{list()}
		@returns: yields L{File} and L{Dir} objects, depth first
		'''
		for name in self.list(raw=raw):
			path = self.path + SEP + name
			if os.path.isdir(path):
				dir = self.subdir(name)
				yield dir
				for child in dir.walk(raw=raw):
					yield child
			else:
				yield self.file(name)

	def get_file_tree_as_text(self, raw=True):
		'''Returns an overview of files and folders below this dir
		as text. Used in tests.
		@param raw: see L{list()}
		@returns: file listing as string
		'''
		text = ''
		for child in self.walk(raw=raw):
			path = child.relpath(self)
			if isinstance(child, Dir):
				path += '/'
			text += path + '\n'
		return text

	def touch(self, mode=None):
		'''Create this folder and any parent folders that do not yet
		exist.
		@param mode: creation mode (e.g. 0700)
		'''
		if self.exists():
			# Additional check needed because makedirs can not handle
			# a path like "E:\" on windows (while "E:\foo" works fine)
			return

		try:
			if mode is not None:
				os.makedirs(self.path, mode=mode)
			else:
				os.makedirs(self.path)
		except OSError as e:
			if e.errno != errno.EEXIST:
				raise

	def remove(self):
		'''Remove this folder, fails if it is not empty.'''
		logger.info('Remove dir: %s', self)
		lrmdir(self.path)
		FS.emit('path-deleted', self)

	def cleanup(self):
		'''Remove this foldder and any empty parent folders. If the
		folder does not exist, still check for empty parent folders.
		Fails silently if the folder is not empty.
		@returns: C{True} when successfull (so C{False} means it still exists).
		'''
		if not self.exists():
			return True

		try:
			os.removedirs(self.path)
		except OSError:
			return False # probably dir not empty
		else:
			return True

	def remove_children(self):
		'''Recursively remove everything below this folder .

		B{WARNING:} This is quite powerful and can do a lot of damage
		when executed for the wrong folder, so pleae make sure to double
		check the dir is actually what you think it is before calling this.
		'''
		assert self.path and self.path != '/'
		logger.info('Remove file tree: %s', self)
		for root, dirs, files in os.walk(self.path, topdown=False):
			# walk should not decent into symlinked folders by default
			# remove() and rmdir() both should remove a symlink rather
			# than the target of the link
			for name in files:
				os.remove(os.path.join(root, name))
			for name in dirs:
				lrmdir(os.path.join(root, name))

	def copyto(self, dest):
		'''Recursively copy the contents of this folder.
		When the destination folder already exists the contents will be
		merged, so you need to check existence of the destination first
		if you want a clean new copy.
		@param dest: a L{Dir} object
		'''
		# We do not use shutil.copytree() because it requires that
		# the target dir does not exist
		assert isinstance(dest, Dir)
		assert not dest == self, 'BUG: trying to copy a dir to itself'
		logger.info('Copy dir %s to %s', self, dest)

		def copy_dir(source, target):
			target.touch()
			for item in source.list():
				child = FilePath((source, item))
				if child.isdir():
					copy_dir(Dir(child), target.subdir(item)) # recur
				else:
					child = File(child)
					child.copyto(target)

		copy_dir(self, dest)
		# TODO - not hooked with FS signals

	def file(self, path):
		'''Get a L{File} object for a path below this folder

		@param path: a (relative) file path as string, tuple or
		L{FilePath} object. When C{path} is a L{File} object already
		this method still enforces it is below this folder.
		So this method can be used as check as well.

		@returns: a L{File} object
		@raises PathLookupError: if the path is not below this folder
		'''
		file = self.resolve_file(path)
		if not file.path.startswith(self.path):
			raise PathLookupError('%s is not below %s' % (file, self))
		return file

	def resolve_file(self, path):
		'''Get a L{File} object for a path relative to this folder

		Like L{file()} but allows the path to start with "../" as
		well, so can handle any relative path.

		@param path: a (relative) file path as string, tuple or
		L{FilePath} object.
		@returns: a L{File} object
		'''
		assert isinstance(path, (FilePath, str, list, tuple))
		if isinstance(path, str):
			return File((self.path, path))
		elif isinstance(path, (list, tuple)):
			return File((self.path,) + tuple(path))
		elif isinstance(path, File):
			return path
		elif isinstance(path, FilePath):
			return File(path.path)

	def new_file(self, path):
		'''Get a L{File} object for a new file below this folder.
		Like L{file()} but guarantees the file does not yet exist by
		adding sequential numbers if needed. So the resulting file
		may have a modified name.

		@param path: a (relative) file path as string, tuple or
		L{FilePath} object.

		@returns: a L{File} object
		@raises PathLookupError: if the path is not below this folder
		'''
		file = self.file(path)
		basename = file.basename
		if '.' in basename:
			basename, ext = basename.split('.', 1)
		else:
			ext = ''
		dir = file.dir
		i = 0
		while file.exists():
			logger.debug('File exists "%s" trying increment', file)
			i += 1
			newname = basename + '%03i' % i
			if ext:
				newname += '.' + ext
			file = dir.file(newname)
		return file

	def subdir(self, path):
		'''Get a L{Dir} object for a path below this folder

		@param path: a (relative) file path as string, tuple or
		L{FilePath} object. When C{path} is a L{Dir} object already
		this method still enforces it is below this folder.
		So this method can be used as check as well.

		@returns: a L{Dir} object
		@raises PathLookupError: if the path is not below this folder

		'''

		dir = self.resolve_dir(path)
		if not dir.path.startswith(self.path):
			raise PathLookupError('%s is not below %s' % (dir, self))
		return dir

	def resolve_dir(self, path):
		'''Get a L{Dir} object for a path relative to this folder

		Like L{subdir()} but allows the path to start with "../" as
		well, so can handle any relative path.

		@param path: a (relative) file path as string, tuple or
		L{FilePath} object.
		@returns: a L{Dir} object
		'''
		assert isinstance(path, (FilePath, str, list, tuple))
		if isinstance(path, str):
			return Dir((self.path, path))
		elif isinstance(path, (list, tuple)):
			return Dir((self.path,) + tuple(path))
		elif isinstance(path, Dir):
			return path
		elif isinstance(path, FilePath):
			return Dir(path.path)

	def new_subdir(self, path):
		'''Get a L{Dir} object for a new sub-folder below this folder.
		Like L{subdir()} but guarantees the folder does not yet exist by
		adding sequential numbers if needed. So the resulting folder
		may have a modified name.

		@param path: a (relative) file path as string, tuple or
		L{FilePath} object.

		@returns: a L{Dir} object
		@raises PathLookupError: if the path is not below this folder
		'''
		subdir = self.subdir(path)
		basename = subdir.basename
		i = 0
		while subdir.exists():
			logger.debug('Dir exists "%s" trying increment', subdir)
			i += 1
			newname = basename + '%03i' % i
			subdir = self.subdir(newname)
		return subdir


def _glob_to_regex(glob):
	glob = glob.replace('.', '\\.')
	glob = glob.replace('*', '.*')
	glob = glob.replace('?', '.?')
	return re.compile(glob)


class FilteredDir(Dir):
	'''Class implementing a folder with a filtered listing. Can be
	used to e.g. filter all objects that are also ignored by version
	control.
	'''

	def __init__(self, path):
		'''Constructor

		@param path: an absolute file path, file URL, L{FilePath} object
		or a list of path elements. When a list is given, the first
		element is allowed to be an absolute path, URL or L{FilePath}
		object as well.
		'''
		Dir.__init__(self, path)
		self._ignore = []

	def ignore(self, glob):
		'''Add a file pattern to ignore
		@param glob: a file path pattern (e.g. "*.txt")
		'''
		regex = _glob_to_regex(glob)
		self._ignore.append(regex)

	def filter(self, name):
		for regex in self._ignore:
			if regex.match(name):
				return False
		else:
			return True

	def list(self, includehidden=False, includetmp=False, raw=False):
		files = Dir.list(self, includehidden, includetmp)
		if not raw:
			files = list(filter(self.filter, files))
		return files


class File(FilePath):
	'''Class representing a single file.

	This class implements much more complex logic than the default
	python file objects. E.g. on writing we first write to a temporary
	files, then flush and sync and finally replace the file we intended
	to write with the temporary file. This makes it much more difficult
	to loose file contents when something goes wrong during the writing.

	Also it implements logic to check the modification time before
	writing to prevent overwriting a file that was changed on disk in
	between read and write operations. If this mtime check fails MD5
	sums are used to verify before raising an exception (because some
	share drives do not maintain mtime very precisely).
	This logic is not atomic, so your mileage may vary.
	'''

	# For atomic write we first write a tmp file which has the extension
	# .zim-new~ when is was written successfully we replace the actual file
	# with the tmp file. Because rename is atomic on POSIX platforms and
	# replaces the existing file this either succeeds or not, it can never
	# truncate the existing file but fail to write the new file. So if writing
	# fails we should always at least have the old file still present.
	# If we encounter a left over .zim-new~ we ignore it since it may be
	# corrupted.
	#
	# For Window the behavior is more complicated, see the WindowsFile class
	# below.
	#
	# Note that the mechanism to avoid overwriting files that changed on disks
	# does not prevent conflicts when two processes try to write to the same
	# file at the same time. This is a hard problem that is currently not
	# addressed in this implementation.

	def __init__(self, path, checkoverwrite=False, endofline=None):
		'''Constructor

		@param path: an absolute file path, file URL, L{FilePath} object
		or a list of path elements. When a list is given, the first
		element is allowed to be an absolute path, URL or L{FilePath}
		object as well.

		@param checkoverwrite: when C{True} this object checks the
		modification time before writing to prevent overwriting a file
		that was changed on disk in between read and write operations.

		@param endofline: the line end style used when writing, can be
		one of "unix" ('\\n') or "dos" ('\\r\\n'). Whan C{None} the local
		default is used.
		'''
		FilePath.__init__(self, path)
		self.checkoverwrite = checkoverwrite
		self.endofline = endofline
		self._mtime = None
		self._md5 = None

	def __eq__(self, other):
		if isinstance(other, File):
			return self.path == other.path
		else:
			return False

	def exists(self):
		return os.path.isfile(self.path)

	def isimage(self):
		'''Check if this is an image file. Convenience method that
		works even when no real mime-type suport is available.
		If this method returns C{True} it is no guarantee
		this image type is actually supported by Gtk.
		@returns: C{True} when this is an image file
		'''

		# Quick shortcut to be able to load images in the gui even if
		# we have no proper mimetype support
		if '.' in self.basename:
			_, ext = self.basename.rsplit('.', 1)
			if ext in IMAGE_EXTENSIONS:
				return True

		return self.get_mimetype().startswith('image/')

	def get_mimetype(self):
		'''Get the mime-type for this file.
		Will use the XDG mimetype system if available, otherwise
		fallsback to the standard library C{mimetypes}.
		@returns: the mimetype as a string, e.g. "text/plain"
		'''
		if xdgmime:
			mimetype = xdgmime.get_type(self.path, name_pri=80)
			return str(mimetype)
		else:
			mimetype, encoding = mimetypes.guess_type(self.path, strict=False)
			if encoding == 'gzip':
				return 'application/x-gzip'
			elif encoding == 'bzip':
				return 'application/x-bzip'
			elif encoding == 'compress':
				return 'application/x-compress'
			else:
				return mimetype or 'application/octet-stream'

	def get_endofline(self):
		'''Get the end-of-line character(s) used for writing this file.
		@returns: the end-of-line character(s)
		'''
		if self.endofline is None:
			if isinstance(self, WindowsPath):
				return '\r\n'
			else:
				return '\n'
		else:
			assert self.endofline in ('unix', 'dos')
			if self.endofline == 'dos':
				return '\r\n'
			else:
				return '\n'

	def raw(self):
		'''Get the raw content without UTF-8 decoding, newline logic,
		etc. Used to read binary data, e.g. when serving files over www.
		Note that this function also does not integrates with checking
		mtime, so intended for read only usage.
		@returns: file content as string
		'''
		try:
			fh = open(self.path, mode='rb')
			content = fh.read()
			fh.close()
			return content
		except IOError:
			raise FileNotFoundError(self)

	def read(self):
		'''Get the file contents as a string. Takes case of decoding
		UTF-8 and fixes line endings.
		@returns: the content as (unicode) string.
		@raises FileNotFoundError: when the file does not exist.
		'''
		try:
			content = self._read()
			self._checkoverwrite(content)
			return content.lstrip('\ufeff').replace('\x00', '')
				# Strip unicode byte order mark
				# And remove any NULL byte since they screw up parsing
		except IOError:
			raise FileNotFoundError(self)
		except UnicodeDecodeError as error:
			raise FileUnicodeError(self, error)

		return text

	def _read(self):
		with open(self.path, encoding='UTF-8') as fh:
			return fh.read()

	def readlines(self):
		'''Get the file contents as a list of lines. Takes case of
		decoding UTF-8 and fixes line endings.

		@returns: the content as a list of lines.
		@raises FileNotFoundError: when the file does not exist.
		'''
		try:
			file = open(self.path, encoding='UTF-8')
			lines = file.readlines()
			self._checkoverwrite(lines)
			return [line.lstrip('\ufeff').replace('\x00', '') for line in lines]
				# Strip unicode byte order mark
				# And remove any NULL byte since they screw up parsing
		except IOError:
			raise FileNotFoundError(self)
		except UnicodeDecodeError as error:
			raise FileUnicodeError(self, error)

		return lines

	def _write_check(self):
		if not self.iswritable():
			raise FileWriteError(_('File is not writable: %s') % self.path) # T: Error message
		elif not self.exists():
			self.dir.touch()
		else:
			pass # exists and writable

	def write(self, text):
		'''Write file contents from string. This overwrites the current
		content. Will automatically create all parent folders.
		If writing fails the file will either have the new content or the
		old content, but it should not be possible to have the content
		truncated.
		@param text: new content as (unicode) string
		@emits: path-created if the file did not yet exist
		'''
		self._assertoverwrite()
		isnew = not os.path.isfile(self.path)
		newline = self.get_endofline()
		self._write_check()
		with AtomicWriteContext(self, newline=newline) as fh:
			fh.write(text)

		self._checkoverwrite(text)
		if isnew:
			FS.emit('path-created', self)

	def writelines(self, lines):
		'''Write file contents from a list of lines.
		Like L{write()} but input is a list instead of a string.
		@param lines: new content as list of lines
		@emits: path-created if the file did not yet exist
		'''
		self._assertoverwrite()
		isnew = not os.path.isfile(self.path)
		newline = self.get_endofline()
		self._write_check()
		with AtomicWriteContext(self, newline=newline) as fh:
			fh.writelines(lines)

		self._checkoverwrite(lines)
		if isnew:
			FS.emit('path-created', self)

	def _checkoverwrite(self, content):
		# Set properties needed by assertoverwrite for the in-memory object
		if self.checkoverwrite:
			self._mtime = self.mtime()
			self._md5 = _md5(content)

	def _assertoverwrite(self):
		# When we read a file and than write it, this method asserts the file
		# did not change in between (e.g. by another process, or another async
		# function of our own process). We use properties of this object instance
		# We check the timestamp, if that does not match we check md5 to be sure.
		# (Sometimes e.g. network filesystems do not maintain timestamps as strict
		# as we would like.)
		#
		# This function should not prohibit writing without reading first.
		# Also we just write the file if it went missing in between
		if self._mtime and self._md5:
			try:
				mtime = self.mtime()
			except OSError:
				if not os.path.isfile(self.path):
					logger.critical('File missing: %s', self.path)
					return
				else:
					raise

			if not self._mtime == mtime:
				logger.warning('mtime check failed for %s, trying md5', self.path)
				if self._md5 != _md5(self._read()):
					raise FileWriteError(_('File changed on disk: %s') % self.path)
						# T: error message
					# Why are we using MD5 here ?? could just compare content...

	def check_has_changed_on_disk(self):
		'''Returns C{True} when this file has changed on disk'''
		if not (self._mtime and self._md5):
			if os.path.isfile(self.path):
				return True # may well been just created
			else:
				return False # ??
		elif not os.path.isfile(self.path):
			return True
		else:
			try:
				self._assertoverwrite()
			except FileWriteError:
				return True
			else:
				return False

	def touch(self):
		'''Create this file and any parent folders if it does not yet
		exist. (Parent folders are also created when writing to a file,
		so you only need to call this method in special cases - e.g.
		when an external program requires the file to exist.)
		'''
		if self.exists():
			return
		else:
			self.write('')

	def remove(self):
		'''Remove (delete) this file and cleanup any related temporary
		files we created. This action can not be un-done.
		Ignores silently if the file did not exist in the first place.
		'''
		logger.info('Remove file: %s', self)
		if os.path.isfile(self.path):
			os.remove(self.path)

		tmp = self.path + '.zim-new~'
		if os.path.isfile(tmp):
			os.remove(tmp)

		FS.emit('path-deleted', self)

	def cleanup(self):
		'''Remove this file and cleanup any empty parent folder.
		Convenience method calling L{File.remove()} and L{Dir.cleanup()}.
		'''
		self.remove()
		self.dir.cleanup()

	def copyto(self, dest):
		'''Copy this file to another location. Preserves all file
		attributes (by using C{shutil.copy2()})
		@param dest: a L{File} or L{Dir} object for the destination. If the
		destination is a folder, we will copy to a file below that
		folder of the same name
		'''
		dest = adapt_from_newfs(dest)
		assert isinstance(dest, (File, Dir))
		if isinstance(dest, Dir):
			assert not dest == self.dir, 'BUG: trying to copy a file to itself'
		else:
			assert not dest == self, 'BUG: trying to copy a file to itself'
		logger.info('Copy %s to %s', self, dest)
		if isinstance(dest, Dir):
			dest.touch()
		else:
			dest.dir.touch()
		shutil.copy2(self.path, dest.path)
		# TODO - not hooked with FS signals

	def compare(self, other):
		'''Check if file contents are the same. This differs from
		L{isequal()} because files can be different physical locations.
		@param other: another L{File} object
		@returns: C{True} when the files have the same content
		'''
		# TODO: can be more efficient, e.g. by checking stat size first
		# also wonder if MD5 is needed here ... could just compare text
		return _md5(self.read()) == _md5(other.read())


class TmpFile(File):
	'''Class for temporary files. These are stored in the temp directory
	and by default they are deleted again when the object is destructed.
	'''

	def __init__(self, basename, unique=True, persistent=False):
		'''Constructor

		@param basename: gives the name for this tmp file.
		@param unique: if C{True} the L{Dir.new_file()} method is used
		to make sure we have a new file.
		@param persistent: if C{False} the file will be removed when the
		object is destructed, if C{True} we leave it alone
		'''
		dir = get_tmpdir()
		if unique:
			file = dir.new_file(basename)
			File.__init__(self, file.path)
		else:
			File.__init__(self, (dir, basename))

		self.persistent = persistent

	def __del__(self):
		if not self.persistent:
			self.remove()



# Replace logic based on discussion here:
# http://stupidpythonideas.blogspot.nl/2014/07/getting-atomic-writes-right.html
#
# The point is to get a function to replace an old file with a new
# file as "atomic" as possible

if hasattr(os, 'replace'):
	_replace_file = os.replace
elif sys.platform == 'win32':
	# The win32api.MoveFileEx method somehow does not like our unicode,
	# the ctypes version does ??!
	import ctypes
	_MoveFileEx = ctypes.windll.kernel32.MoveFileExW
	_MoveFileEx.argtypes = [ctypes.c_wchar_p, ctypes.c_wchar_p, ctypes.c_uint32]
	_MoveFileEx.restype = ctypes.c_bool
	def _replace_file(src, dst):
		try:
			if not _MoveFileEx(src, dst, 1): # MOVEFILE_REPLACE_EXISTING
				raise OSError('Could not replace "%s" -> "%s"' % (src, dst))
		except:
			# Sometimes it fails - we play stupid and try again...
			time.sleep(0.5)
			if not _MoveFileEx(src, dst, 1): # MOVEFILE_REPLACE_EXISTING
				raise OSError('Could not replace "%s" -> "%s"' % (src, dst))
else:
	_replace_file = os.rename


### TODO filter Dir.list directly for hidden files
if os.name != 'nt':
	def is_hidden_file(file):
			return file.basename.startswith('.')

else:
	import ctypes

	def is_hidden_file(file):
		INVALID_FILE_ATTRIBUTES = -1
		FILE_ATTRIBUTE_HIDDEN = 2

		try:
			attrs = ctypes.windll.kernel32.GetFileAttributesW(file.path)
				# note: GetFileAttributesW is unicode version of GetFileAttributes
		except AttributeError:
			return False
		else:
			if attrs == INVALID_FILE_ATTRIBUTES:
				return False
			else:
				return bool(attrs & FILE_ATTRIBUTE_HIDDEN)
###


class FSObjectMonitor(SignalEmitter):

	__signals__ = {
		'changed': (None, None, (None, None)),
	}

	def __init__(self, path):
		self.path = path
		self._gio_file_monitor = None

	def _setup_signal(self, signal):
		if signal == 'changed' \
		and self._gio_file_monitor is None \
		and Gio:
			try:
				file = Gio.File.new_for_uri(self.path.uri)
				self._gio_file_monitor = file.monitor()
				self._gio_file_monitor.connect('changed', self._on_changed)
			except:
				logger.exception('Error while setting up file monitor')

	def _teardown_signal(self, signal):
		if signal == 'changed' \
		and self._gio_file_monitor:
			try:
				self._gio_file_monitor.cancel()
			except:
				logger.exception('Error while tearing down file monitor')
			finally:
				self._gio_file_monitor = None

	def _on_changed(self, filemonitor, file, other_file, event_type):
		# 'FILE_MONITOR_EVENT_CHANGED' is always followed by
		# a 'FILE_MONITOR_EVENT_CHANGES_DONE_HINT' when the filehandle
		# is closed (or after timeout). Idem for "created", assuming it
		# is not created empty.
		#
		# TODO: do not emit changed on CREATED - separate signal that
		#       can be used when monitoring a file list, but reserve
		#       changed for changes-done-hint so that we ensure the
		#       content is complete.
		#       + emit on write and block redundant signals here
		#
		# Also note that in many cases "moved" will not be used, but a
		# sequence of deleted, created will be signaled
		#
		# For Dir objects, the event will refer to files contained in
		# the dir.

		#~ print('MONITOR:', self, event_type)
		if event_type in (
			Gio.FileMonitorEvent.CREATED,
			Gio.FileMonitorEvent.CHANGES_DONE_HINT,
			Gio.FileMonitorEvent.DELETED,
			Gio.FileMonitorEvent.MOVED,
		):
			self.emit('changed', None, None) # TODO translate otherfile and eventtype




# Copyright 2011 Marcel Stimberg <stimberg@users.sourceforge.net>

'''Push events to the Zeitgeist daemon'''

import logging
import sys

from gi.repository import Gio

from zim.plugins import PluginClass
from zim.notebook import NotebookExtension
from zim.signals import SIGNAL_AFTER
from zim.newfs import LocalFile
from zim.fs import adapt_from_oldfs

from zim.gui.mainwindow import MainWindowExtension


logger = logging.getLogger('zim.plugins.zeitgeist')


try:
	from zeitgeist.client import ZeitgeistClient
	from zeitgeist.datamodel import Event, Subject, Interpretation, Manifestation
except:
	ZeitgeistClient = None


class ZeitgeistPlugin(PluginClass):

	plugin_info = {
		'name': _('Log events with Zeitgeist'), # T: plugin name
		'description': _('Pushes events to the Zeitgeist daemon.'), # T: plugin description
		'author': 'Marcel Stimberg',
		'help': 'Plugins:Log events with Zeitgeist',
	}

	@classmethod
	def check_dependencies(klass):
		has_zeitgeist = not ZeitgeistClient is None
		return has_zeitgeist, [('libzeitgeist', has_zeitgeist, False)]

	def __init__(self):
		PluginClass.__init__(self)
		try:
			self.zeitgeist_client = ZeitgeistClient()
			self.zeitgeist_client.register_data_source('application://zim.desktop',
			    'Zim', _('Zim Desktop Wiki'), []) # T: short description of zim
		except RuntimeError as e:
			logger.exception('Loading zeitgeist client failed, will not log events')
			self.zeitgeist_client = None

	def create_and_send_event(self, page, event_type):
		if not self.zeitgeist_client:
			return

		source_file = adapt_from_oldfs(page.source_file)
		if not isinstance(source_file, LocalFile):
			return

		uri = page.source_file.uri
		origin = page.source_file.dirname
		text = _('Wiki page: %s') % page.name
			# T: label for how zim pages show up in the recent files menu, %s is the page name

		subject = Subject.new_for_values(mimetype='text/x-zim-wiki',
		                                 uri=uri,
		                                 origin=origin,
		                                 interpretation=Interpretation.PLAIN_TEXT_DOCUMENT,
		                                 manifestation=Manifestation.FILE_DATA_OBJECT,
		                                 text=text)
		event = Event.new_for_values(actor='application://zim.desktop',
		                             interpretation=event_type,
		                             manifestation=Manifestation.USER_ACTIVITY,
		                             subjects=[subject, ])

		self.zeitgeist_client.insert_event(event)


class ZeitGeistMainWindowExtension(MainWindowExtension):

	def __init__(self, plugin, window):
		MainWindowExtension.__init__(self, plugin, window)
		self.connectto(window, 'page-changed')
		if window.page is not None:
			self.plugin.create_and_send_event(window.page, Interpretation.ACCESS_EVENT)
			self.page = window.page
		else:
			self.page = None

	def on_page_changed(self, pageview, page):
		if self.page is not None:
			self.plugin.create_and_send_event(self.page, Interpretation.LEAVE_EVENT)

		self.plugin.create_and_send_event(page, Interpretation.ACCESS_EVENT)
		self.page = page

	def destroy(self):
		if self.page is not None:
			self.plugin.create_and_send_event(self.page, Interpretation.LEAVE_EVENT)
			self.page = None


class ZeitgeistNotebookExtension(NotebookExtension):

	def __init__(self, plugin, notebook):
		NotebookExtension.__init__(self, plugin, notebook)
		self.connectto_all(notebook,
			('deleted-page', 'stored-page'), order=SIGNAL_AFTER)

	def on_deleted_page(self, notebook, page):
		logger.debug("Deleted page: %s", page.name)
		self.plugin.create_and_send_event(page, Interpretation.DELETE_EVENT)

	def on_stored_page(self, notebook, page):
		logger.debug("Modified page: %s", page.name)
		self.plugin.create_and_send_event(page, Interpretation.MODIFY_EVENT)



# Copyright 2011 Jaap Karssenberg <jaap.karssenberg@gmail.com>

class Builder(object):
	'''This class defines a 'builder' interface.
	It is used by parsers to construct the parse tree while keeping the
	parser objects agnostic of how the resulting parse tree objects
	look.
	'''

	def start(self, tag, attrib=None):
		'''Start formatted region
		@param tag: the tag name
		@param attrib: optional dict with attributes
		@implementation: must be implemented by sub-classes
		'''
		raise NotImplemented

	def text(self, text):
		'''Append text
		@param text: text to be appended as string
		@implementation: must be implemented by sub-classes
		'''
		raise NotImplemented

	def end(self, tag):
		'''End formatted region
		@param tag: the tag name
		@raises XXX: when tag does not match current state
		@implementation: must be implemented by sub-classes
		'''
		raise NotImplemented

	def append(self, tag, attrib=None, text=None):
		'''Convenience function to open a tag, append text and close
		it immediatly. Only used for formatted text that has no
		sub-processing done.
		@param tag: the tag name
		@param attrib: optional dict with attributes
		@param text: formatted text
		@implementation: optional for subclasses, default implementation
		calls L{start()}, L{text()}, and L{end()}
		'''
		self.start(tag, attrib)
		if not text is None:
			self.text(text)
		self.end(tag)


class BuilderTextBuffer(Builder):
	'''Wrapper that buffers text going to a L{Builder} object
	such that the last piece of text remains accessible for inspection
	and can be modified.
	'''

	def __init__(self, builder):
		self.builder = builder
		self.buffer = []

	# Interface to handle text buffer

	def get_text(self):
		return ''.join(self.buffer)

	def set_text(self, text):
		self.buffer = [text]

	def clear_text(self):
		self.buffer = []

	def flush(self):
		text = ''.join(self.buffer)
		if text:
			self.builder.text(text)
		self.buffer = []

	# Builder interface

	def start(self, tag, attrib=None):
		if self.buffer:
			self.flush()
		self.builder.start(tag, attrib)

	def end(self, tag):
		if self.buffer:
			self.flush()
		self.builder.end(tag)

	def text(self, text):
		self.buffer.append(text)

	def append(self, tag, attrib=None, text=None):
		if self.buffer:
			self.flush()
		self.builder.append(tag, attrib, text)



# Copyright 2008-2014 Jaap Karssenberg <jaap.karssenberg@gmail.com>

'''This module contains the main object to "execute" a template and
fill in the parameters, call functions etc. The L{TemplateProcessor}
defined here takes care of the template control flow ('IF', 'FOR', etc.).

Also see the L{expression} sub module that contains logic for executing
expressions in the template.
'''


import collections.abc as abc

from zim.newfs import FileNotFoundError
from zim.base import MovingWindowIter
from zim.parse.simpletree import SimpleTreeElement

from zim.templates.expression import ExpressionDictObject, ExpressionParameter


class TemplateContextDict(ExpressionDictObject):
	'''This class defines a dict with template parameters

	These dicts can be modified when running the template, but nested
	dicts can only be modified if they are a TemplateContextDict
	themselves.
	'''

	_fmethods = ExpressionDictObject._fmethods + (
		'pop', 'clear', 'update', 'setdefault', 'items'
	) # adding methods for mutable mapping here


class TemplateProcessor(object):
	'''The template processor takes a parsed template and "executes" it
	one or more times.
	'''

	# See Expression for remarks on safe eval of expressions.
	#
	# In addition here we set also parameters in the context rather
	# then just retrieving them. To make this safe we only allow
	# modification of C{TemplateContextDict} dicts and do not allow
	# any assignment to other dicts or objects.

	# Instructions supported here:
	#
	# 	'GET', 'SET',
	# 	'IF', 'ELIF', 'ELSE',
	# 	'FOR'
	# 	'INCLUDE',

	def __init__(self, parts, parse_included_file_func=None):
		'''Constructor
		@param parts: A list of L{SimplerTreeElements} as produced by
		L{TemplateParser.parse()}
		'''
		self.main = None
		self.blocks = {}
		self.parse_included_file_func = parse_included_file_func
		for item in parts:
			if item.tag == 'MAIN':
				self.main = item
			elif item.tag == 'BLOCK':
				self.blocks[item.get('name')] = item
			else:
				raise AssertionError('Unknown tag: %s' % item.tag)

		if self.main is None:
			raise AssertionError('Missing main part of template')

	def process(self, output, context):
		'''Execute the template once
		@param output: an object to receive the template output, can be
		a C{list} and should support at least an C{append()} method to
		receive string content
		@param context: a L{TemplateContextDict} object with the
		template parameters
		'''
		assert isinstance(context, TemplateContextDict)
		self.__call__(output, self.main, context)

	@staticmethod
	def _set(context, var, value):
		# We only allow setting in pre-defined TemplateContextDict's
		namespace = context
		for name in var.parts[:-1]:
			if namespace \
			and isinstance(namespace, TemplateContextDict):
				namespace = namespace.get(name)
			else:
				raise AssertionError('Can not assign: %s' % var.name)

		if namespace is not None \
		and isinstance(namespace, TemplateContextDict):
			namespace[var.key] = value
		else:
			raise AssertionError('Can not assign: %s' % var.name)

	def __call__(self, output, elements, context):
		n = len(elements)
		i = 0
		while i < n:
			try:
				element = elements[i]
				i += 1
				if isinstance(element, str):
					output.append(element)
				elif element.tag == 'GET':
					expr = element.attrib['expr']
					value = expr(context)
					output.append(str(value))
				elif element.tag == 'SET':
					var = element.attrib['var']
					expr = element.attrib['expr']
					value = expr(context)
					self._set(context, var, value)
				elif element.tag in ('IF', 'ELIF'):
					expr = element.attrib['expr']
					if bool(expr(context)):
						self.__call__(output, element, context) # recurs
						while i < n \
						and isinstance(elements[i], SimpleTreeElement) \
						and elements[i].tag in ('ELIF', 'ELSE'):
							# Skip subsequent ELIF / ELSE clauses
							i += 1
				elif element.tag == 'ELSE':
					self.__call__(output, element, context) # recurs
				elif element.tag == 'FOR':
					self._loop(output, element, context)
				elif element.tag == 'INCLUDE':
					expr = element.attrib['expr']
					if isinstance(expr, ExpressionParameter) and expr.name in self.blocks:
						# INCLUDE name - do not evaluate as an expression
						self._include_block(output, expr.name, context)
					else:
						# INCLUDE expr - eval to either block name or file path
						arg = expr(context)
						if '/' in arg or '\\' in arg or '.' in arg:
							self._include_file(output, arg, context)
						else:
							self._include_block(output, arg, context)
				else:
					raise AssertionError('Unknown instruction: %s' % element.tag)
			except:
				raise

	def _loop(self, output, element, context):
		var = element.attrib['var']
		expr = element.attrib['expr']
		items = expr(context)
		if not isinstance(items, abc.Iterable):
			raise TypeError('Can not iterate over: %s' % items)
		elif not isinstance(items, abc.Sized):
			# cast to list to ensure we have a len()
			items = list(items)

		# set "loop"
		outer = context.get('loop')
		if isinstance(outer, TemplateLoopState):
			loop = TemplateLoopState(len(items), outer)
		else:
			loop = TemplateLoopState(len(items), None)
		context['loop'] = loop

		# do the iterations
		myiter = MovingWindowIter(items)
		for i, items in enumerate(myiter):
			loop._update(i, myiter)
			self._set(context, var, items[1]) # set var
			self.__call__(output, element, context) # recurs

		# restore "loop"
		context['loop'] = outer

	def _include_block(self, output, name, context):
		if name in self.blocks:
			self.__call__(output, self.blocks[name], context) # recurs
		else:
			raise AssertionError('No such block defined: %s' % name)

	def _include_file(self, output, path, context):
		if self.parse_included_file_func is None:
			raise AssertionError('No template resources provided')

		try:
			include_template = None
			for item in self.parse_included_file_func(path):
				if item.tag == 'MAIN':
					include_template = item
				elif item.tag == 'BLOCK':
					self.blocks[item.get('name')] = item
				else:
					raise AssertionError('Unknown tag: %s' % item.tag)
		except FileNotFoundError:
			raise AssertionError('No such file in template resources: %s' % path)
		else:
			if include_template is None:
				raise AssertionError('BUG: Error while parsing INCLUDE (!?)')
			else:
				self.__call__(output, include_template, context) # recurs


class TemplateLoopState(object):
	'''Object used for the "loop" parameter in a FOR loop'''

	def __init__(self, size=None, outer=None):
		if size is None:
			self.size = None
			self.max = None
		else:
			self.size = size
			self.max = size - 1

		self.outer = outer

		self.prev = None
		self.current = None
		self.next = None
		self.index = None
		self.count = None
		self.first = None
		self.last = None
		self.parity = None
		self.even = None
		self.odd = None

	def _update(self, i, movingwindowiter):
		self.prev = movingwindowiter.items[0]
		self.current = movingwindowiter.items[1]
		self.next = movingwindowiter.items[2]

		self.first = i == 0
		self.last = movingwindowiter.last

		self.index = i
		self.count = i + 1
		self.parity = 'even' if i % 2 == 0 else 'odd'
		self.even = self.parity == 'even'
		self.odd = self.parity == 'odd'



import ntpath
import os
import pathlib

import gi

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

from zim.gui.widgets import Dialog, InputEntry
from zim.newfs import localFileOrFolder, LocalFile, LocalFolder


class FileRenameDialog(Dialog):
	""" A dialog for renaming a file. """

	def __init__(self, parent, file):
		title = _('Rename file') # T: dialog title
		Dialog.__init__(self, parent, title)

		assert isinstance(file, LocalFile) and not isinstance(file, LocalFolder)
		self.old_file = file
		self.new_file = file

		# Add field for entering new filename.
		self.txt_filename = InputEntry()
		self.txt_filename.set_text(self.old_file.basename)
		self.txt_filename.set_activates_default(True)  # Make ENTER key press trigger the OK button.
		self.txt_filename.connect("changed", self.do_validate, parent)

		# Add field for showing hints when an error occurs (e.g. filename already exists).
		self.txt_error = Gtk.Label()
		self.txt_error.set_visible(False)

		# Add ok button.
		self.btn_ok = self.get_widget_for_response(response_id=Gtk.ResponseType.OK)
		self.btn_ok.set_can_default(True)
		self.btn_ok.grab_default()
		self.btn_ok.set_sensitive(False)

		# Configure dialog.
		self.set_modal(True)
		self.set_default_size(380, 100)
		self.vbox.pack_start(self.txt_filename, False, True, 0)
		self.vbox.pack_start(self.txt_error, False, True, 0)

		# Set focus to search field
		self.txt_filename.grab_focus()

	def do_validate(self, widget, data):
		""" Validating new file name, show error when validation fails and enable/disable ok button. """

		def is_filename(filename):
			""" Returns True when filename does not contain any path declaration. """
			return filename == os.path.basename(filename)

		def does_file_already_exist(filename):
			""" Checks whether the new filename is already taken. """
			return self.old_file.parent().file(filename).exists()

		def show_error(msg):
			""" Displays an error and disables the OK button. """
			self.txt_error.set_text(msg)
			self.txt_error.set_visible(True)
			self.btn_ok.set_sensitive(False)
			return False

		if not widget.get_text():
			return show_error(_('File name should not be blank.')) # T: Error message

		if does_file_already_exist(widget.get_text()):
			return show_error(_('A file with that name already exists.')) # T: Error message

		if not is_filename(widget.get_text()):
			return show_error(_('File name should not contain path declaration.')) # T: Error message

		# No errors, hide error label and enable OK button.
		self.txt_error.hide()
		self.btn_ok.set_sensitive(True)
		return True

	def do_response_ok(self):
		self.new_file = self.old_file.parent().file(self.txt_filename.get_text())
		self.result = Gtk.ResponseType.OK
		self.close()




# Copyright 2011 Greg Warner <gdwarner@gmail.com>
# (Pretty much copied from diagrameditor.py)

import logging

from zim.plugins import PluginClass
from zim.plugins.base.imagegenerator import \
	ImageGeneratorClass, BackwardImageGeneratorObjectType

from zim.newfs import LocalFile, TmpFile
from zim.applications import Application, ApplicationError
from zim.gui.base.images import supports_image_format

logger = logging.getLogger('zim.plugins.sequencediagrameditor')

def get_cmd(fmt):
	return ('seqdiag', '-T', fmt, '-o')


class InsertSequenceDiagramPlugin(PluginClass):

	plugin_info = {
		'name': _('Insert Sequence Diagram'), # T: plugin name
		'description': _('''\
This plugin provides a sequence diagram editor for zim based on seqdiag.
It allows easy editing of sequence diagrams.
'''), # T: plugin description
		'help': 'Plugins:Sequence Diagram Editor',
		'author': 'Greg Warner',
	}

	plugin_preferences = (
		# key, type, label, default
		(
			'prefer_svg',
			'bool',
            _('Generate diagrams in SVG format'), # T: plugin preference
			supports_image_format('svg')
		),
	)

	@classmethod
	def check_dependencies(klass):
		has_diagcmd = Application(get_cmd('png')).tryexec()
		return has_diagcmd, [("seqdiag", has_diagcmd, True)]


class BackwardSequenceDiagramImageObjectType(BackwardImageGeneratorObjectType):

	name = 'image+seqdiagram'
	label = _('Sequence Diagram') # T: menu item
	syntax = None
	scriptname = 'seqdiagram.diag'


class SequenceDiagramGenerator(ImageGeneratorClass):
	@property
	def _pref_format(self):
		return 'svg' if self.plugin.preferences['prefer_svg'] else 'png'

	@property
	def imagefile_extension(self):
		return '.' + self._pref_format

	@property
	def diagcmd(self):
		return get_cmd(self._pref_format)

	def __init__(self, plugin, notebook, page):
		ImageGeneratorClass.__init__(self, plugin, notebook, page)
		self.diagfile = TmpFile('seqdiagram.diag')
		self.diagfile.touch()

	def generate_image(self, text):
		# Write to tmp file
		self.diagfile.write(text)
		self.imgfile = LocalFile(self.diagfile.path[:-5] + self.imagefile_extension) # len('.diag') == 5
		logger.debug('Writing diagram to temp file: %s', self.imgfile)

		# Call seqdiag
		try:
			diag = Application(self.diagcmd)
			diag.run((self.imgfile, self.diagfile))
		except ApplicationError:
			return None, None # Sorry, no log
		else:
			return self.imgfile, None

	def cleanup(self):
		self.diagfile.remove()
		try:
			self.imgfile.remove()
		except AttributeError:
			logger.debug('Closed dialog before generating image, nothing to remove')




# Copyright 2008-2014 Jaap Karssenberg <jaap.karssenberg@gmail.com>

'''This module contains the framework for exporting data from zim.

The main API for exporting from the application is the L{Exporter}
object. There are subclasses of Exporter to export to multiple files,
to a single file or to a MHTML file.

To configure the exporter object an additional L{ExportLayout} object
is used. This layout defines the exact mapping of pages to files to
be used.

To specific the pages to export a subclass of the L{PageSelection} class
is used. There are selection classes to export the whole notebook or
to export a single page.

The L{ExportTemplateContext} object defines specific template parameters
for exporting. See also L{zim.templates} for template parsing and
processing classes.

See the module functions for convenient standard configurations.
'''


# TODO - when exporting with namespace / prefix we should also trim
#        links within a SingleFile output relative to that prefix
#        --> do not leak info on parent namespace names



from zim.fs import adapt_from_oldfs
from zim.newfs import LocalFolder, LocalFile
from zim.templates import get_template
from zim.formats import get_format


def build_notebook_exporter(dir, format, template, **opts):
	'''Returns an L{Exporter} that is suitable for exporting a whole
	notebook to a folder with one file per page
	'''
	from zim.export.layouts import MultiFileLayout
	from zim.export.exporters.files import MultiFileExporter

	dir = adapt_from_oldfs(dir)

	template = get_template(format, template)
	ext = get_format(format).info['extension']
	layout = MultiFileLayout(dir, ext)
	return MultiFileExporter(layout, template, format, **opts)


def build_page_exporter(file, format, template, page, **opts):
	'''Returns an L{Exporter} that is suitable for exporting a page with
	subpages to a file and a folder (e.g. "page.html" with "page_files/")
	'''
	from zim.export.layouts import FileLayout
	from zim.export.exporters.files import MultiFileExporter

	file = adapt_from_oldfs(file)

	template = get_template(format, template)
	ext = get_format(format).info['extension']
	layout = FileLayout(file, page, ext)
	return MultiFileExporter(layout, template, format, **opts)


def build_single_file_exporter(file, format, template, namespace=None, **opts):
	'''Returns an L{Exporter} that is suitable for exporting a set of
	pages to a single file
	'''
	from zim.export.layouts import SingleFileLayout
	from zim.export.exporters.files import SingleFileExporter

	file = adapt_from_oldfs(file)

	template = get_template(format, template)
	layout = SingleFileLayout(file)
	return SingleFileExporter(layout, template, format, **opts)


def build_mhtml_file_exporter(file, template, **opts):
	'''Returns an L{Exporter} that is suitable for exporting a set of
	pages to a single mhtml file
	'''
	from zim.export.exporters.mhtml import MHTMLExporter

	file = adapt_from_oldfs(file)

	template = get_template('html', template)
	return MHTMLExporter(file, template, **opts)




# Copyright 2017 Jaap Karssenberg <jaap.karssenberg@gmail.com>

'''
Functions to match and parse dates in wiki pages

The following unambigous date forms are supported:
* Day by date using: yyyy-mm-dd  for example 2017-02-16
* Month using: yyyy-mm for example 2017-02
* Week using: (yy)yyWww or (yy)yy-Www, for example 2017W07, 17W07 or 17-W07
* Day by week using  (yy)yyWww(D) or (yy)yy-Www(-D) for example 17-W07-2 for Tuesday
* Week and day by week notation can also use Wkyyww(.D) for example wk1707

To avoid confusion between mm/dd, dd/mm and yy-mm notations neither of
these is supported and the year should always be given in 4 digits.

When the year is shortened to two digits ("yy") it is always prefixed by "20",
so "01" becomes 2001 and "99" becomes 2099. For years starting with 19 (or any
other century) the ful four digit year needs to be used.

Notation is based on https://en.wikipedia.org/wiki/ISO_8601. Truncating the year
is not supported except for the week notation and separators ("-") are required
to avoid matching random other numbers. The "Wkyyww" notation is supported as
non-standard alternative.

Weeknumbers follow the iso calendar. However depending on locale Sunday can
either be the first day of the starting week or the last day of the ending week.
In the weekday notation this is made explicit by using "0" for sunday at the
start of the week and "7" for sunday at the end of the week. Thus the dates
"W1707.7" and "W1708.0" are the same day.
'''

# TODO: add alternative notation using month abbr "16 FEB 2017"


import re
import datetime

from zim.datetimetz import dates_for_week, weekcalendar


__all__ = ('date_re', 'parse_date', 'Month', 'Week', 'Day')


date_re = re.compile(
	r'(?:'
	r'\d{4}-\d{2}-\d{2}'
	r'|\d{4}-\d{2}'
	r'|(?:\d{2}|\d{4})-?[Ww][Kk]?\d{2}(?:-\d)?'
	r'|[Ww][Kk]?(?:\d{2}|\d{4})\d{2}(?:[.-]\d)?'
	')(?![\\w-])'
)


def parse_date(date):
	string = date.upper().replace('-', '').strip()
	if 'W' in string:
		string = string.replace('WK', '').replace('W', '').replace('.', '')
		if len(string) == 4: # yyww
			return Week(int(string[:2]) + 2000, int(string[2:4]))
		elif len(string) == 5: # yywwD
			return Day.new_from_weeknumber(int(string[:2]) + 2000, int(string[2:4]), int(string[4]))
		elif len(string) == 6: # yyyyww
			return Week(int(string[:4]), int(string[4:]))
		elif len(string) == 7: # yyyywwD
			return Day.new_from_weeknumber(int(string[:4]), int(string[4:6]), int(string[6]))
		else:
			raise ValueError('Could not parse: %s' % date)
	elif len(string) == 6: # yyyymm
		return Month(int(string[:4]), int(string[4:]))
	elif len(string) == 8: # yyyymmdd
		return Day(int(string[:4]), int(string[4:6]), int(string[6:]))
	else:
		raise ValueError('Could not parse: %s' % date)


class DateRange(object):

	def __repr__(self):
		return "<%s: %s>" % (self.__class__.__name__, str(self))


class Day(DateRange, datetime.date):

	@classmethod
	def new_from_weeknumber(cls, year, week, weekday):
		if not (isinstance(weekday, int) and 0 <= weekday <= 7):
			raise ValueError('Not a weekday: %i (must be between 0 and 7)' % weekday)

		start, end = dates_for_week(year, week)
		if start.isoweekday() == 1: # monday
			offset = weekday - 1
		else: # sunday
			offset = weekday

		if offset != 0:
			start = start + datetime.timedelta(days=offset)
		return cls(start.year, start.month, start.day)

	@property
	def first_day(self):
		return self

	@property
	def last_day(self):
		return self

	def weekcalendar(self):
		'''Returns (year, week, weekday)'''
		year, week, weekday = weekcalendar(self)
		if weekday == 1 and self.isoweekday() == 7:
			weekday = 0 # See module doc on weekday
		else:
			weekday = self.isoweekday()
		return year, week, weekday

	def weekformat(self):
		'''Format as iso-weeknumber and weekday "YYYY-Www-D"'''
		return '%s-W%s-%s' % self.weekcalendar()


class Week(DateRange):

	def __init__(self, year, week):
		self.year = year
		self.week = week
		self.first_day, self.last_day = dates_for_week(year, week)

	def __str__(self):
		return '%s-W%s' % (self.year, self.week)


class Month(DateRange):

	def __init__(self, year, month):
		self.year = year
		self.month = month
		self.first_day = datetime.date(year, month, 1)
			# ensures year and month are actually valid

	@property
	def last_day(self):
		if self.month < 12:
			return datetime.date(self.year, self.month + 1, 1) - datetime.timedelta(days=1)
		else:
			return datetime.date(self.year + 1, 1, 1) - datetime.timedelta(days=1)

	def __str__(self):
		return '%s-%s' % (self.year, self.month)


def old_parse_date(string):
	'''Returns a tuple of (year, month, day) for a date string or None
	
	NOTE: only included for backward compatibility, do not use in new code
	
	Current supported formats:

		- C{dd?-mm?}
		- C{dd?-mm?-yy}
		- C{dd?-mm?-yyyy}
		- C{yyyy-mm?-dd?}

	Where '-' can be replaced by any separator. Any preceding or
	trailing text will be ignored (so we can parse journal page names
	correctly).

	TODO: Some setting to prefer US dates with mm-dd instead of dd-mm
	TODO: More date formats ?
	'''
	m = re.search(r'(\d{1,4})\D(\d{1,2})(?:\D(\d{1,4}))?', string)
	if m:
		d, m, y = m.groups()
		if len(d) == 4:
			y, m, d = d, m, y
		if not d:
			return None # yyyy-mm not supported

		if not y:
			# Guess year, based on time delta
			from datetime import date
			today = date.today()
			if today.month - int(m) >= 6:
				y = today.year + 1
			else:
				y = today.year
		else:
			y = int(y)
			if y < 50:
				y += 2000
			elif y < 1000:
				y += 1900

		return tuple(map(int, (y, m, d)))
	else:
		return None



# Copyright 2008-2023 Jaap Karssenberg <jaap.karssenberg@gmail.com>

import weakref

from gi.repository import GObject
from gi.repository import Gtk

from zim.gui.insertedobjects import InsertedObjectWidget

from .constants import LINE, OBJECT
from .find import PluginInsertedObjectFindMixin


class InsertedObjectAnchor(Gtk.TextChildAnchor):

	is_inline = False # T: whether or not this object can be inline in a paragraph

	def create_widget(self):
		raise NotImplementedError

	def dump(self, builder):
		raise NotImplementedError


class LineSeparatorAnchor(InsertedObjectAnchor):

	def __init__(self):
		GObject.GObject.__init__(self)
		self.objectmodel = None

	def create_widget(self):
		return LineSeparator()

	def dump(self, builder):
		builder.start(LINE, {})
		builder.data('-'*20) # FIXME: get rid of text here
		builder.end(LINE)


class LineSeparator(InsertedObjectWidget):
	'''Class to create a separation line.'''

	def __init__(self):
		InsertedObjectWidget.__init__(self)
		widget = Gtk.Box()
		widget.get_style_context().add_class(Gtk.STYLE_CLASS_BACKGROUND)
		widget.set_size_request(-1, 3)
		self.add(widget)


class TableAnchor(PluginInsertedObjectFindMixin, InsertedObjectAnchor):
	# HACK - table support is native in formats, but widget is still in plugin
	#        so we need to "glue" the table tokens to the plugin widget

	def __init__(self, objecttype, objectmodel):
		GObject.GObject.__init__(self)
		self.objecttype = objecttype
		self.objectmodel = objectmodel
		self.widgets = weakref.WeakSet()

	def create_widget(self):
		widget = self.objecttype.create_widget(self.objectmodel)
		self.widgets.add(widget)
		return widget

	def dump(self, builder):
		self.objecttype.dump(builder, self.objectmodel)


class PluginInsertedObjectAnchor(PluginInsertedObjectFindMixin, InsertedObjectAnchor):

	def __init__(self, objecttype, objectmodel):
		GObject.GObject.__init__(self)
		self.objecttype = objecttype
		self.objectmodel = objectmodel
		self.is_inline = objecttype.is_inline
		self.widgets = weakref.WeakSet()

	def create_widget(self):
		widget = self.objecttype.create_widget(self.objectmodel)
		self.widgets.add(widget)
		return widget

	def dump(self, builder):
		attrib, data = self.objecttype.data_from_model(self.objectmodel)
		builder.start(OBJECT, dict(attrib)) # dict() because ElementTree doesn't like ConfigDict
		if data is not None:
			builder.data(data)
		builder.end(OBJECT)





# Copyright 2008-2014 Jaap Karssenberg <jaap.karssenberg@gmail.com>

import email.mime.multipart
import email.mime.text
import email.mime.nonmultipart

import base64

from zim.fs import adapt_from_oldfs
from zim.newfs import LocalFile, LocalFolder, File, Folder, get_tmpdir

from zim.notebook import encode_filename

from zim.export.exporters import Exporter
from zim.export.exporters.files import SingleFileExporter
from zim.export.layouts import SingleFileLayout
from zim.export.linker import ExportLinker


class MHTMLExporter(Exporter):
	'''Exporter that exports pages and attachments to a single MHTML
	file.
	'''

	# Output is Multipart-Mime message containing single HTML file
	# and all attachments and resources as mime parts
	# So first export as a single file, then wrap in mime

	# Keeps large data in memory - would need more low level
	# implementation with incremental writes to optimize it...

	# Also note that due to all the base64 encoding, size is going
	# to blow up even more ...

	def __init__(self, file, template, document_root_url=None):
		self.file = file
		self.template = template
		self.document_root_url = document_root_url

	def export_iter(self, pages):
		basename = encode_filename(pages.name)
		folder = get_tmpdir().folder('mhtml_export_tmp_dir')
		if folder.exists():
			folder.remove_children()
		else:
			folder.touch()
		file = folder.file(basename + '.html')
		layout = SingleFileLayout(file, pages.prefix)
		exporter = SingleFileExporter(layout, self.template, 'html', document_root_url=self.document_root_url)

		for p in exporter.export_iter(pages):
			yield p

		encoder = MHTMLEncoder()
		linker = ExportLinker(pages.notebook, layout, output=file, usebase=True)
		self.file.write(encoder(layout, linker))


class MHTMLEncoder(object):

	# Create message of file + attachments + resources

	# We use a linker for relative names to make absolutely sure
	# we give same relative paths as mentioned in links

	def __call__(self, layout, linker):
		msg = email.mime.multipart.MIMEMultipart('related')
			# MIME-Version 1.0
			# Content-Type: multipart/related; boundry=...
		msg.preamble = '' \
		'This document is a Single File Web Page, also known as a Web Archive file\n' \
		'or MHTML. If you are seeing this message, your browser or editor doesn\'t\n' \
		'support MHTML. Please look for a plugin or extension that adds MHTML support\n' \
		'or download a browser that supports it.'

		# Add html file
		msg.attach(self.encode_text_file(layout.file, None, 'text/html'))

		# Add attachments and resource
		for file in self._walk(layout.dir):
			file = adapt_from_oldfs(file)
			mt = file.mimetype()
			filename = linker.link(file.uri)
			if mt.startswith('text/'):
				part = self.encode_text_file(file, filename, mt)
			else:
				part = self.encode_data_file(file, filename, mt)
			msg.attach(part)

		# Write message to file
		return str(msg)

	def _walk(self, dir):
		if dir.exists():
			for child in dir:
				if isinstance(child, File):
					yield child
				elif child.exists():
					for child in self._walk(child): # recurs
						yield child

	def encode_text_file(self, file, filename, mimetype):
		type, subtype = mimetype.split('/', 1)
		assert type == 'text'

		# Not using MIMEText here, since it uses base64 and inflates
		# all ascii text unnecessary
		charset = email.charset.Charset('utf-8')
		charset.body_encoding = email.charset.QP
		msg = email.mime.nonmultipart.MIMENonMultipart('text', subtype, charset='utf-8')
		if filename: # top level does not have filename
			msg['Content-Location'] = filename
		msg.set_payload(file.read(), charset=charset)
		return msg

	def encode_data_file(self, file, filename, mimetype):
		file = adapt_from_oldfs(file)
		type, subtype = mimetype.split('/', 1)
		msg = email.mime.nonmultipart.MIMENonMultipart(type, subtype)
		msg['Content-Location'] = filename
		msg['Content-Type'] = mimetype
		msg.set_payload(file.read_binary())
		email.encoders.encode_base64(msg)
		return msg




# Copyright 2009-2013 Jaap Karssenberg <jaap.karssenberg@gmail.com>

# The Error class needed to be put in a separate file to avoid recursive
# imports.

'''This module contains the base class for all errors in zim'''

import sys
import logging

logger = logging.getLogger('zim')


use_gtk_errordialog = False

def set_use_gtk(use_gtk):
	'''Set whether or not L{show_error} and L{exception_handler}
	shold use the L{ErrorDialog} or not.
	@param use_gtk: set C{True} for interactive gui, C{False} for
	terminal mode
	'''
	global use_gtk_errordialog
	use_gtk_errordialog = use_gtk


def get_error_msg(error):
	'''Returns the message to show for an error
	@param error: error object or string
	@returns: 2-tuple of: message string and a boolean
	whether a traceback should be shown or not
	'''
	if isinstance(error, Error):
		# An "expected" error
		return error.msg, False

	elif isinstance(error, EnvironmentError):
		# Normal error, e.g. OSError or IOError
		msg = error.strerror
		if hasattr(error, 'filename') and error.filename:
			msg += ': ' + error.filename
		return msg, False

	else:
		# An unexpected error, all other Exception's
		msg = _('Looks like you found a bug') # T: generic error dialog
		return msg, True


def log_error(error, debug=None):
	'''Log error and traceback
	@param error: error as understood by L{get_error_msg()}
	@param debug: optional debug message, defaults to the error itself
	'''
	msg, show_trace = get_error_msg(error)
	if debug is None:
		debug = msg

	if show_trace:
		# unexpected error - will be logged with traceback
		logger.exception(debug)
	else:
		# expected error - log trace to debug
		logger.debug(debug, exc_info=1)
		logger.error(msg)


def _run_error_dialog(error):
	#~ try:
	from zim.gui.widgets import ErrorDialog
	ErrorDialog(None, error, do_logging=False).run()
	#~ except:
		#~ logger.error('Failed to run error dialog')


def show_error(error):
	'''Show an error by calling L{log_error()} and when running
	interactive also calling L{ErrorDialog}.
	@param error: the error object
	'''
	log_error(error)
	if use_gtk_errordialog:
		_run_error_dialog(error)


def exception_handler(debug):
	'''Like C{show_error()} but with debug message instead of the actual
	error. Intended to be used in C{except} blocks as a catch-all for
	both intended and unintended errors.
	@param debug: debug message for logging
	'''
	# We use debug as log message, rather than the error itself
	# the error itself shows up in the traceback anyway
	exc_info = sys.exc_info()
	error = exc_info[1]
	del exc_info # recommended by manual

	log_error(error, debug=debug)
	if use_gtk_errordialog:
		_run_error_dialog(error)


class Error(Exception):
	'''Base class for all errors in zim.

	This class is intended for application and usage errors, these will
	be caught in the user interface and presented as error dialogs.
	In contrast and Exception that does I{not} derive from this base
	class will result in a "You found a bug" dialog. Do not use this
	class e.g. to catch programming errors.

	Subclasses should define two attributes. The first is 'msg', which is
	the short description of the error. Typically this gives the specific
	input / page / ... which caused the error. In there should be an attribute
	'description' (either as class attribute or object attribute) with a verbose
	description. This description can be less specific but should explain
	the error in a user friendly way. The default behavior is to take 'msg' as
	the single argument for the constructor. So a minimal subclass only needs
	to define a class attribute 'description'.

	For a typical error dialog in the Gtk interface the short string from 'msg'
	will be shown as the title in bold letters while the longer 'description'
	is shown below it in normal letters. As a guideline error classes that are
	used in the gui or that can be e.g. be raised on invalid input from the
	user should be translated.
	'''

	description = ''
	msg = '<Unknown Error>'
		# in case subclass does not define instance attribute

	def __init__(self, msg, description=None):
		self.msg = msg
		if description:
			self.description = description
			# else use class attribute

	def __str__(self):
		msg = '' + self.msg.strip()
		if self.description:
			msg += '\n\n' + self.description.strip() + '\n'
		return msg

	def __repr__(self):
		return '<%s: %s>' % (self.__class__.__name__, self.msg)




# Copyright 2020-2023 Jaap Karssenberg <jaap.karssenberg@gmail.com>

# This module contains functions to import text files into a zim
# notebook

# TODO:
# - implement import_folder
# - autodetect format to import from
# - import attachments as well as pages (incl. filter to force e.g. txt as attachment)


import logging

from zim.newfs import File
from zim.notebook.page import Path


logger = logging.getLogger('zim')


def _page_name_from_file(file):
	name = file.basename
	if '.' in name:
		name, e = name.rsplit('.', 1) # remove extension, .txt / .md / ...
	return Path.makeValidPageName(name)


def import_file_from_user_input(file, notebook, path=None, format='wiki'):
	'''Wrapper for L{import_file()} which handles sanatizing user input
	Imports a single file or folder to a given notebook location
	@param file: file or folder object to be imported
	@notebook: notebook object
	@path: **target** notebook path, if C{None} import in top-level
	@returns: imported page object
	'''
	if path is None:
		path = Path(_page_name_from_file(file))
	else:
		path = notebook.pages.lookup_from_user_input(path.name)

	return import_file(file, notebook, path)


def import_files_from_user_input(files, notebook, path=None, format='wiki'):
	'''Wrapper for L{import_file() and L{import_folder()} which handles sanatizing user input
	Imports multiple files or folders as children of a given notebook location
	@param files: sequence of file or folder objects to be imported
	@notebook: notebook object
	@path: **parent** notebook path, if C{None} import into the top-level
	'''
	if path is None:
		path = Path(':')
	else:
		path = notebook.pages.lookup_from_user_input(path.name)

	for f in files:
		p = path + _page_name_from_file(f)
		if isinstance(f, File):
			import_file(f, notebook, p, format)
		else:
			import_folder(f, notebook, p, format)


def import_file(file, notebook, path, format='wiki'):
	'''Import a file into a zim notebook page
	@param file: a L{File} object to import from
	@param notebook: a L{Notebook} object to import into
	@param path: the L{Path} to import to within the notebook
	@returns: the L{Page} object for the imported page, this may be a different
	page than the one C{path} is pointing - see L{notebook.get_new_page()}
	'''
	logging.debug('Import file "%s" to "%s" as %s', file, path, format)
	if not file.istext:
		raise ValueError('Not a text file: %s' % file.path)

	if file.ischild(notebook.folder):
		newfile = file.parent().new_file(file.basename + '~')
		file.moveto(newfile)
		file = newfile

	page = notebook.get_new_page(path)
	assert not page.exists()

	page.parse(format, file.readlines())
	notebook.store_page(page)
	return page


def import_folder(folder, notebook, path, filter=None, format='wiki'):
	'''Import a folder recursively
	@param folder: a L{Folder} object to import from
	@param notebook: a L{Notebook} object to import into
	@param path: the L{Path} to import to within the notebook
	@param filter: optional filter function to decide which files and folder
	to import when scanning C{folder}. Will be given each file or folder found
	when scanning C{folder} and expected to return boolean whether to import
	yes or no.
	'''
	# Preserve folder / sub-folder names when creating target paths
	raise NotImplementedError



# Copyright 2009-2024 Jaap Karssenberg <jaap.karssenberg@gmail.com>

'''This module contains various classes and functions to help with parsing text content
and handling tokenlist or parse-tree structures.

These are utilities that are used in various parts of the application.

See also the various sub-modules
'''

import re



def convert_space_to_tab(text, tabstop=4):
	'''Convert spaces to tabs
	@param text: the input text
	@param tabstop: the number of spaces to represent a tab
	@returns: the fixed text
	'''
	# Fix tabs
	spaces = ' ' * tabstop
	pattern = '^(\t*)((?:%s)+)' % spaces
	text = re.sub(
		pattern,
		lambda m: m.group(1) + '\t' * (len(m.group(2)) // tabstop),
		text,
		flags=re.M
	)
	return text


def fix_unicode_whitespace(text):
	'''Replaces unicode whitespace characters
	These characters are recognized by "splitlines()" but not as end-of-line
	in regexes. See also issue #1760
	@param text: the input text
	@returns: the fixed text
	'''
	text = text.replace('\u2028', '\n') # LINE SEPARATOR
	text = text.replace('\u2029', ' ') # PARAGRAPH SEPARATOR
	return text




# Copyright 2008-2014 Jaap Karssenberg <jaap.karssenberg@gmail.com>


'''PageSelection objects wrap the list of pages to be exported'''


from zim.notebook import PageNotFoundError


class PageSelection(object):
	'''Base class defining the public API'''

	notebook = None #: The L{Notebook} object
	prefix = None #: optional L{Path} object with common prefix of pages in the selection, or C{None}
	name = None #: name used in export template
	title = None # name used in export template

	def __iter__(self):
		'''Iterate page objects
		@implementation: must be implemented by subclases
		'''
		raise NotImplemented

	def index(self, namespace=None):
		'''Iterate path objects for template C{index()} function, depth first
		@param namespace: the sub namespace to iterate or None to iterate
		toplevel
		'''
		raise NotImplemented

	# TODO add __len__ that gives total pages for progress
	# TODO Use collections subclass to make interface complete ?


class AllPages(PageSelection):
	'''Selection of all pages in a notebook'''

	def __init__(self, notebook):
		self.notebook = notebook
		self.name = notebook.name
		self.title = notebook.name # XXX implement notebook.title

	def __iter__(self):
		for path in self.notebook.pages.walk():
			try:
				yield self.notebook.get_page(path)
			except PageNotFoundError:
				pass

	def index(self, namespace=None):
		return self.notebook.pages.walk(namespace)


class SinglePage(PageSelection):
	'''Selection of one specific page without subpages'''

	def __init__(self, notebook, page):
		self.notebook = notebook
		self.page = page
		self.name = self.page.name
		self.title = self.page.name # XXX implement page.title (use heading)
		self.prefix = page

	def __iter__(self):
		try:
			yield self.notebook.get_page(self.page)
		except PageNotFoundError:
			pass

	def index(self, namespace=None):
		if namespace is None or self.page.ischild(namespace):
			yield self.page
		else:
			pass


class SubPages(SinglePage):
	'''Selection of pages in sub-tree of a notebook'''

	def __iter__(self):
		yield self.notebook.get_page(self.page)
		for path in self.notebook.pages.walk(self.page):
			yield self.notebook.get_page(path)

	def index(self, namespace=None):
		if namespace is None or namespace.name == self.page.name:
			yield self.page
			for page in self.notebook.pages.walk(self.page):
				yield page
		elif namespace.ischild(self.page):
			for page in self.notebook.pages.walk(namespace):
				yield page
		else:
			pass



# Copyright 2011 Patricio Paez <pp@pp.com.mx>
#
# Plugin to use arithmetic in Zim wiki

from zim.inc.arithmetic import ParserGTK

from zim.plugins import PluginClass
from zim.actions import action

from zim.gui.pageview import PageViewExtension


class ArithmeticPlugin(PluginClass):

	plugin_info = {
		'name': _('Arithmetic'), # T: plugin name
		'description': _('''\
This plugin allows you to embed arithmetic calculations in zim.
It is based on the arithmetic module from
http://pp.com.mx/python/arithmetic.
'''), # T: plugin description
		'author': 'Patricio Paez',
		'help': 'Plugins:Arithmetic',
	}

	plugin_preferences = (
		# key, type, label, default
		('output_decimals', 'int', _('Output precision'), 6, (0, 100)), # T: plugin preference
	)


class ArithmeticPageViewExtension(PageViewExtension):

	def __init__(self, plugin, window):
		PageViewExtension.__init__(self, plugin, window)
		self.preferences = plugin.preferences

	@action(_('_Arithmetic'), accelerator='F5', icon='accessories-calculator', menuhints='tools') # T: menu item
	def calculate(self):
		"""Perform arithmetic operations"""

		# get the buffer
		buf = self.pageview.textview.get_buffer() # XXX

		# parse and return modified text
		parser = ParserGTK(output_decimals=self.preferences['output_decimals'])
		parser.parse(buf)



#
# scoreeditor.py
#
# This is a plugin for Zim, which allows to insert music score in zim using
# GNU Lilypond.
#
#
# Author: Shoban Preeth <shoban.preeth@gmail.com>
# Date: 2012-07-05
# Copyright (c) 2012, released under the GNU GPL v2 or higher
#
#

import glob

from zim.plugins import PluginClass
from zim.plugins.base.imagegenerator import ImageGeneratorClass, BackwardImageGeneratorObjectType
from zim.newfs import LocalFile, TmpFile
from zim.config import data_file
from zim.templates import get_template
from zim.applications import Application, ApplicationError


# TODO put these commands in preferences
lilypond_cmd = ('lilypond', '-ddelete-intermediate-files',
		# '-dsafe', # Can't include files in safe mode
		'-dbackend=eps', '--png', '--header=texidoc')
convertly_cmd = ('convert-ly', '--current-version', '--edit')
lilypondver_cmd = ('lilypond', '--version')


def _get_lilypond_version():
	try:
		lilypond = Application(lilypondver_cmd)
		output = lilypond.pipe()
		return output[0].split()[2]
	except ApplicationError:
		return '2.14.2'

class InsertScorePlugin(PluginClass):

	plugin_info = {
		'name': _('Insert Score'), # T: plugin name
		'description': _('''\
This plugin provides an score editor for zim based on GNU Lilypond.

This is a core plugin shipping with zim.
'''), # T: plugin description
		'help': 'Plugins:Score Editor',
		'author': 'Shoban Preeth',
	}

	plugin_preferences = [
		# key, type, label, default
		('include_header', 'string', _('Common include header'), '\\include "predefined-guitar-fretboards.ly"'), # T: plugin preference
		('include_footer', 'string', _('Common include footer'), ''), # T: plugin preference
	]

	@classmethod
	def check_dependencies(klass):
		has_lilypond = Application(lilypond_cmd).tryexec()
		has_convertly = Application(convertly_cmd).tryexec()
		return has_lilypond and has_convertly, [
				('GNU Lilypond', has_lilypond, True),
				('convert-ly', has_convertly, True)
			]


class BackwardScoreImageObjectType(BackwardImageGeneratorObjectType):

	name = 'image+score'
	label = _('Musical Score') # T: menu item
	syntax = None
	scriptname = 'score.ly'


class ScoreGenerator(ImageGeneratorClass):

	imagefile_extension = '.png'
	
	cur_lilypond_version = None

	def __init__(self, plugin, notebook, page):
		ImageGeneratorClass.__init__(self, plugin, notebook, page)
		self.template = get_template('plugins', 'scoreeditor.ly')
		self.scorefile = TmpFile('score.ly')
		self.cur_lilypond_version = _get_lilypond_version()
		self.include_header = plugin.preferences['include_header']
		self.include_footer = plugin.preferences['include_footer']

	def check_user_input(self, text):
		version_present = False
		for l in text.splitlines(True):
			if l.strip().startswith('\\version'):
				version_present = True
		if not version_present:
			text = '\\version "{0}"\n\n'.format(self.cur_lilypond_version) + text
		return text

	def extract_version(self, text):
		outtext = []
		version = None
		for l in text.splitlines(True):
			if l.strip().startswith('\\version'):
				version = l.strip()
			else:
				outtext.append(l)
		return (version, ''.join(outtext))

	def generate_image(self, text):
		(version, text) = self.extract_version(text)
		#~ print('>>>%s<<<' % text)

		# Write to tmp file using the template for the header / footer
		scorefile = self.scorefile
		lines = []
		self.template.process(lines, {
			'score': text,
			'version': version or '',
			'include_header': self.include_header or '',
			'include_footer': self.include_footer or '',
		})
		scorefile.writelines(lines)
		#~ print('>>>%s<<<' % scorefile.read())

		# Call convert-ly to convert document of current version of
		# Lilypond.
		clogfile = LocalFile(scorefile.path[:-3] + '-convertly.log') # len('.ly) == 3
		try:
			convertly = Application(convertly_cmd)
			convertly.run((scorefile.basename,), cwd=scorefile.parent())
		except ApplicationError:
			clogfile.write('convert-ly failed.\n')
			return None, clogfile


		# Call lilypond to generate image.
		logfile = LocalFile(scorefile.path[:-3] + '.log') # len('.ly') == 3
		try:
			lilypond = Application(lilypond_cmd)
			lilypond.run(('-dlog-file=' + logfile.basename[:-4], scorefile.basename,), cwd=scorefile.parent())
		except ApplicationError:
			# log should have details of failure
			return None, logfile
		pngfile = LocalFile(scorefile.path[:-3] + '.png') # len('.ly') == 3

		return pngfile, logfile

	def cleanup(self):
		path = self.scorefile.path
		for path in glob.glob(path[:-3] + '*'):
			LocalFile(path).remove()




# Copyright 2008-2018 Jaap Karssenberg <jaap.karssenberg@gmail.com>

import logging

logger = logging.getLogger('zim.gui')


from gi.repository import Gtk

from zim.plugins import PluginManager
from zim.gui.widgets import Dialog, get_window, InputForm
from zim.parse.links import is_interwiki_keyword_re

notebook_properties = (
	('name', 'string', _('Name')), # T: label for properties dialog
	('interwiki', 'string', _('Interwiki Keyword'), lambda v: not v or is_interwiki_keyword_re.search(v)), # T: label for properties dialog
	('home', 'page', _('Home Page')), # T: label for properties dialog
	('icon', 'image', _('Icon')), # T: label for properties dialog
	('document_root', 'dir', _('Document Root')), # T: label for properties dialog
	('short_links', 'bool', _('Prefer short names for page links'), False), # T: label for properties dialog
	('disable_trash', 'bool', _('Do not use system trash for this notebook'), False), # T: label for properties dialog
	('paste_image_template', 'string', _('Filename template for pasted images')),
	# 'shared' property is not shown in properties anymore
)


class PropertiesDialog(Dialog):

	def __init__(self, parent, notebook, chosen_plugin=None):
		Dialog.__init__(self, parent, _('Properties'), help='Help:Properties') # T: Dialog title
		self.notebook = notebook

		stack = Gtk.Stack()
		sidebar = Gtk.StackSidebar()
		sidebar.set_stack(stack)

		hbox = Gtk.Box()
		hbox.add(sidebar)
		hbox.add(stack)
		self.vbox.add(hbox)

		def add_widget(form, name, title):
			if chosen_plugin and chosen_plugin != name:
				return
			if self.notebook.readonly:
				for widget in list(form.widgets.values()):
					widget.set_sensitive(False)
			box = Gtk.VBox()
			box.pack_start(form, False, False, 0)
			stack.add_titled(box, name, title)

		self.form = InputForm(
			inputs=notebook_properties,
			values=notebook.config['Notebook']
		)
		self.form.widgets['icon'].set_use_relative_paths(self.notebook)
		self.form.widgets['document_root'].set_use_relative_paths(self.notebook)
		add_widget(self.form, 'notebook', _('Notebook'))

		self.plugin_forms = {}
		plugins = PluginManager()
		for name in plugins:
			plugin = plugins[name]
			if plugin.plugin_notebook_properties:
				key = plugin.config_key
				form = InputForm(
					inputs=plugin.form_fields(plugin.plugin_notebook_properties),
					values=notebook.config[key]
				)
				self.plugin_forms[key] = form
				add_widget(form, name, plugin.plugin_info['name'])

	def do_response_ok(self):
		if not self.notebook.readonly:
			properties = self.form.copy()
			properties['icon'] = self.form.widgets['icon'].get_text() # XXX should be file, but resolves relative
			properties['document_root'] = self.form.widgets['document_root'].get_text() # XXX should be file, but resolves relative

			self.notebook.properties.update(properties)

			for key, form in self.plugin_forms.items():
				self.notebook.config[key].update(form)

			if hasattr(self.notebook.config, 'write'): # XXX Check needed for tests
				logger.debug('Write notebook properties')
				self.notebook.config.write()

		return True

## TODO: put a number of properties in an expander with a label "Advanced"




# Copyright 2009-2012 Jaap Karssenberg <jaap.karssenberg@gmail.com>
# Copyright 2012 Damien Accorsi <damien.accorsi@free.fr>



import os
import logging

import xml.etree.ElementTree # needed to compile with cElementTree
try:
	import xml.etree.cElementTree as ET
except:  #pragma: no cover
	import xml.etree.ElementTree as ET


from zim.newfs import LocalFile
from zim.plugins.versioncontrol import VCSApplicationBase
from zim.applications import Application


logger = logging.getLogger('zim.vcs.hg')


class HGApplicationBackend(VCSApplicationBase):

	@classmethod
	def build_bin_application_instance(cls):
		return Application(('hg', '--noninteractive'))
			# force hg to run in non-interactive mode
			# which will force user name to be auto-setup

	def get_mandatory_params(self):
		return ['--noninteractive'] # force hg to run in non-interactive mode
									# which will force user name to be auto-setup

	def build_revision_arguments(self, versions):
		"""Build a list including required string/int for running an VCS command
		# Accepts: None, int, string, (int,), (int, int)
		# Always returns a list
		# versions content:
		  - None: return an empty list
		  - int ou string: return ['-r', int]
		  - tuple or list: return ['-r', '%i..%i']

		It's all based on the fact that defining revision with current VCS is:
		-r revision
		-r rev1..rev2
		"""
		if isinstance(versions, (tuple, list)):
			assert 1 <= len(versions) <= 2
			if len(versions) == 2:
				versions = sorted(map(int, versions))
				return ['-r', '%i..%i' % tuple(versions)]
			else:
				versions = versions[0]

		if not versions is None:
			version = int(versions)
			return ['-r', '%i' % version]
		else:
			return []

	########
	#
	# NOW ARE ALL REVISION CONTROL SYSTEM SHORTCUTS

	def add(self, path=None):
		"""
		Runs: hg add {{PATH}}
		"""
		if path is None:
			return self.run(['add', self.notebook_dir])
		else:
			return self.run(['add', path])


	def annotate(self, file, version=None):
		"""FIXME Document
		return
		0: line1
		2: line1
		...
		"""
		revision_args = self.build_revision_arguments(version)
		return self.pipe(['annotate', file] + revision_args)

	def cat(self, path, version):
		"""
		Runs: hg cat {{PATH}} {{REV_ARGS}}
		"""
		revision_args = self.build_revision_arguments(version)
		return self.pipe(['cat', path] + revision_args)

	def commit(self, path, msg):
		"""
		Runs: hg commit -m {{MSG}} {{PATH}}
		"""
		params = ['commit']
		if msg != '' and msg is not None:
			params.append('-m')
			params.append(msg)
		if path != '' and path is not None:
			params.append(path)
		return self.run(params)

	def diff(self, versions=None, file=None):
		"""
		Runs:
			hg diff --git {{REVISION_ARGS}}
		or
			hg diff --git {{REVISION_ARGS}} {{PATH}}
		"""
		revision_args = self.build_revision_arguments(versions)
		if file is None:
			return self.pipe(['diff', '--git'] + revision_args)
			# Using --git option allow to show the renaming of files
		else:
			return self.pipe(['diff', '--git', file] + revision_args)

	def ignore(self, file_to_ignore_regexp):
		"""
		Build a .hgignore file including the file_to_ignore_content
		@param file_to_ignore_regexp: str representing the .hgignore file content.
		       this must be a list of regexp defining the file / path to ignore,
		       separated by a\n char
		@returns: nothing
		"""
		#TODO: append the rule instead of overwrite the full content
		self.root.file('.hgignore').write(file_to_ignore_regexp)


	def init_repo(self):
		"""Initialize a new repo
		The init operation consists in:
		- running the VCS init command
		- defining files to ignore
		- adding all other existing files
		@returns: nothing
		"""
		self.init()
		self.ignore(r'\.zim*$\n')
		self.add('.') # add all existing files

	def repo_exists(self):
		"""Returns True if a repository is already setup, or False

		@returns: a boolean True if a repo is already setup, or False
		"""
		return self.root.folder('.hg').exists()

	def init(self):
		"""
		Runs: hg init
		"""
		return self.run(['init'])

	def is_modified(self):
		"""Returns true if the repo is not up-to-date, or False
		@returns: True if the repo is not up-to-date, or False
		"""
		# If status return an empty answer, this means the local repo is up-to-date
		return ''.join(self.status()).strip() != ''

	def log(self, path=None):
		"""
		Runs: hg log --style xml {{PATH}}
		"""
		if path:
			return self.pipe(['log', '--style', 'xml', path])
		else:
			return self.pipe(['log', '--style', 'xml'])

	def log_to_revision_list(self, log_op_output):
		# returns a list of tuple (revision-id, date, user, commit-message)
		versions = []
		xml = ET.fromstring(''.join(log_op_output))
		if not (xml and xml.tag == 'log'):
			raise AssertionError('Could not parse log')
		for entry in xml:
			rev = entry.attrib['revision']
			date = entry.findtext('date')
			user = entry.findtext('author')
			msg = entry.findtext('msg')
			versions.append((rev, date, user, msg))

		versions.sort()
		return versions

	def move(self, oldpath, newpath):
		"""
		Runs: hg mv --after {{OLDPATH}} {{NEWPATH}}
		"""
		return self.run(['mv', '--after', oldpath, newpath])

	def remove(self, path):
		"""
		Runs: hg rm {{PATH}}
		"""
		if isinstance(path, LocalFile):
			return self.run(['rm', '-A', path])

	def revert(self, path=None, version=None):
		"""
		Runs:
			hg revert --no-backup {{PATH}} {{REV_ARGS}}
		or
			hg revert --no-backup --all {{REV_ARGS}}
		"""
		revision_params = self.build_revision_arguments(version)
		if path:
			return self.run(['revert', '--no-backup', path] + revision_params)
		else:
			return self.run(['revert', '--no-backup', '--all'] + revision_params)

	def status(self):
		"""
		Runs: hg status
		"""
		return self.pipe(['status'])



#
# ditaaeditor.py
#
# This is a plugin for Zim, which allows coverting ASCII art into proper bitmap
# graphics.
#
#
# Author: Yao-Po Wang <blue119@gmail.com>
# Date: 2012-03-11
# Copyright (c) 2012, released under the GNU GPL v2 or higher
#
#

from zim.plugins import PluginClass
from zim.plugins.base.imagegenerator import ImageGeneratorClass, BackwardImageGeneratorObjectType
from zim.newfs import LocalFile, TmpFile
from zim.config import data_file
from zim.applications import Application, ApplicationError


# TODO put these commands in preferences
dotcmd = ('ditaa')


class InsertDitaaPlugin(PluginClass):

	plugin_info = {
		'name': _('Insert Ditaa'), # T: plugin name
		'description': _('''\
This plugin provides a diagram editor for zim based on Ditaa.

This is a core plugin shipping with zim.
'''), # T: plugin description
        'help': 'Plugins:Ditaa Editor',
		'author': 'Yao-Po Wang',
	}

	@classmethod
	def check_dependencies(klass):
		has_dotcmd = Application(dotcmd).tryexec()
		return has_dotcmd, [("Ditaa", has_dotcmd, True)]


class BackwardDitaaImageObjectType(BackwardImageGeneratorObjectType):

	name = 'image+ditaa'
	label = _('Ascii graph (Ditaa)') # T: menu item
	syntax = None
	scriptname = 'ditaa.dia'


class DitaaGenerator(ImageGeneratorClass):

	imagefile_extension = '.png'

	def __init__(self, plugin, notebook, page):
		ImageGeneratorClass.__init__(self, plugin, notebook, page)
		self.dotfile = TmpFile('ditaa.dia')
		self.dotfile.touch()
		self.pngfile = LocalFile(self.dotfile.path[:-4] + '.png') # len('.dia') == 4

	def generate_image(self, text):
		# Write to tmp file
		self.dotfile.write(text)

		# Call GraphViz
		try:
			dot = Application(dotcmd)
			dot.run((self.dotfile, '-o', self.pngfile))
		except ApplicationError:
			return None, None # Sorry, no log
		else:
			return self.pngfile, None

	def cleanup(self):
		self.dotfile.remove()
		self.pngfile.remove()
